!===============================================================
!
! Copyright (C) 2005 Finite Difference Research Group
!           (C) 2009 Northwestern University, Weizmann Institute of Science
! This file is part of parsec, http://www.ices.utexas.edu/parsec/
! And copyright (C) 2011-2022 Baruch Feldman, with modifications for Transec 
!
! These subroutines implement complex absorbing potentials (CAPs) 
! for use in conductivity/conductance calculations.
!
!---------------------------------------------------------------

! COND - this entire file


!---------------------------------------------------------------
! The main conductance subroutine (also known as TRANSEC :).  
! Adds complex absorbing potentials (CAPs) Gamma to the Kohn-Sham Hamiltonian H_KS 
! from PARSEC and rediagonalizes H_KS + i Gamma, then computes the transmission spectrum T(E).
!---------------------------------------------------------------
subroutine postSCF_conductance(cond,parallel,elec_st,ipr,ierr,&
      grid, pot, solver, symm, clust)
      use constants
      use parallel_data_module
      use conductance_module
      use electronic_struct_module
      use grid_module
      use potential_module
      use eigen_solver_module
      use symmetry_module
      use cluster_module
      
#ifdef MPI
      use mpi
#endif

      implicit none

!!!!!!!!!!!!!!!!!!!!!!!     
!     Input/Output Variables
!     
!     The grid points and mappings
      type(grid_data), intent(in) :: grid
!	Parameters defining the conductance calculation to perform 
      type(conductance_param), intent(inout) :: cond
!     The parallel environment
      type(parallel_data), intent(in) :: parallel
!     Electronic structure information
      type(electronic_struct), intent(inout) :: elec_st
!	Potential information
      type(potential), intent(in) :: pot
!	Eigensolver information
	type(eigen_solver), intent(inout) :: solver
!	Symmetry information
      type(symmetry), intent(in) :: symm
!	Atomic information
      type(cluster), intent(in) :: clust
!     Print-out flag for ARPACK  
      integer, intent(in) :: ipr
!     Error flag from ARPACK 
      integer, intent(out) :: ierr

#ifdef MPI
! Control parameters for asynchronous MPI communication
      integer mpinfo, masterid, group_masterid, comm, group_comm
#endif

!!!!!!!!!!!!!!!!!!!!!!!!!     
!     Work variables
!     
	! Iteration indices (over energies, spin, and grid points), and error status variable for allocations
      integer :: i, k, isp, kplp, igrid, jgrid, kgrid, alcstat
#ifdef MPI
      integer mpistatus(MPI_STATUS_SIZE)
#endif
	  ! If a periodic calculation without R_z,pi symmetry, which sign(s) of current k-point should be performed on current MPI group  
	  integer :: SignK, iEigensolution 
	  integer, parameter :: NEGATIVE_K = -1, BOTH_K = 0, POSITIVE_K = +1, SYMM_K = +2, NO_K = -10 
      ! Number of states to request and maximum ARPACK iterations for eigensolution of H + i Gamma
      integer :: nstate, nec, maxmvs
      ! Work vars for forming CAPs
      real(dp) :: lstdev2, rstdev2, h, zz, r2
	! Parameters describing the Fermi energy and energy spectrum for T(E) calculation
      real(dp) :: MuL, MuR, TWidth, Emin, Emax, dE 
      ! Intermediate and final values for computation of T(E)
      complex(dpc) :: tau  
	! Atomic coordinates and rotation matrix for checking symmetry in periodic boundary condition (PBC) calculations  
	  real(dp), allocatable :: coorat(:, :)  
	  integer, allocatable :: ity(:) 
	  real(dp) :: r(49, 3, 3), trans(3, 3, 48), tnp(48, 3) 
	! Index to symmetry ops while checking symmetry in PBC calculations
	  integer :: ib(48), na, li, ntrans, invers_no
	! Put date stamp after TRANSEC calculation
      real(dp) :: tBegin, tBeginK, tEnd    
      character(len=26) :: datelabel
      ! Spin identifier for filenames, output
      character(len=25) :: spinID 
	! File unit to be used for writing eigenpairs, transformed CAPs, etc.  Note the file units 
	! are chosen to avoid simultaneous writes to the same unit by different MPI groups
	integer :: fileUnit
	! Work var signifying the current k-pt is non-zero, i.e. not the Gamma point 
	logical :: kpt_PBC 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
!     Work arrays for conductance.f90
!     
      ! To hold the CAPs in the diagonal basis of H + i Gamma
      complex(dpc), dimension (:,:), allocatable ::  GamLR 
      ! The full grid of CAPs (gathered from all processors) for writing to disk  
      real(dp), dimension(parallel%ndim) :: gammalall, gammarall
      ! For storing computed eigenvalues & eigenvectors of H + i Gamma
      complex(dpc), allocatable, dimension(:) :: eval, evalNoRz 
      complex(dpc), allocatable, dimension(:, :) :: evec, evecNoRz 
	  integer, allocatable, dimension(:) :: bestMatchInd
	! Work var for use by ARPACK during eigensolution 
      complex(dpc), dimension(parallel%ldn) :: resid_restart  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
!     External function
      intrinsic DABS, exp

!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
!     Interface for routine with optional parameters 
	interface
		subroutine orthonormalize_GramSchmidt_writeEvecs(evec, eval, nstate, cond, parallel, grid, kpt_PBC, kplp, Hks_cplx, gammafull, &
#ifdef MPI 
			mpinfo, & 
#endif
			spinID, evecNoRz, evalNoRz, bestMatchInd)
			use constants
			use conductance_module
			use parallel_data_module
			use grid_module
			integer, intent(in) :: nstate 
			type(conductance_param), intent(in) :: cond
			type(parallel_data), intent(in) :: parallel
			type(grid_data), intent(in) :: grid
			integer, intent(in) :: kplp 
			logical, intent(in) :: kpt_PBC, Hks_cplx 
			real(dp), intent(in)  :: gammafull(parallel%ldn)
			complex(dpc), intent(in) :: eval(nstate)
			complex(dpc), intent(inout) :: evec(parallel%ldn, nstate)
			complex(dpc), intent(in), optional :: evalNoRz(nstate)
			complex(dpc), intent(inout), optional :: evecNoRz(parallel%ldn, nstate) 
#ifdef MPI
			integer, intent(inout) :: mpinfo   
#endif
			character(len=25), optional, intent(in) :: spinID 
			integer, intent(in), optional :: bestMatchInd (nstate)
		end subroutine orthonormalize_GramSchmidt_writeEvecs
	end interface 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!         
!     Memory Allocations
!     

      ! Allocate diagonal part of left CAP GammaL
      allocate(cond%gammal(parallel%ldn),stat=alcstat)
      call alccheck('cond%gammal', parallel%ldn, alcstat)
	! Allocate diagonal part of right CAP GammaR
      allocate(cond%gammar(parallel%ldn),stat=alcstat)
      call alccheck('cond%gammar', parallel%ldn, alcstat)
	! Allocate diagonal part of total CAP, GammaL+GammaR
      allocate(cond%gammafull(parallel%ldn),stat=alcstat)
      call alccheck('cond%gammafull', parallel%ldn, alcstat)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!         
#ifdef MPI
! Define master and comm environment
      masterid = parallel%masterid
      group_masterid = parallel%group_master
      comm = parallel%comm
      group_comm = parallel%group_comm

! Broadcast CAP parameters to all processors
write(9,*) ' conductance(): Before Barrier '
call myflush(9)
      call MPI_BARRIER(comm,mpinfo)
write(9,*) ' conductance(): Before Bcasts '
call myflush(9)


      call MPI_BCAST(cond%gaussians(1)%stdev,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      write(9,*) ' conductance(): After gaussians%stdev Bcast '
	call myflush(9)
	call MPI_BARRIER(comm,mpinfo)
      call MPI_BCAST(cond%gaussians(1)%height,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      call MPI_BCAST(cond%gaussians(1)%z0,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      call MPI_BCAST(cond%gaussians(2)%stdev,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      call MPI_BCAST(cond%gaussians(2)%height,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      call MPI_BCAST(cond%gaussians(2)%z0,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      write(9,*) ' conductance(): After gaussians%z0 Bcast '
        call myflush(9)
        call MPI_BARRIER(comm,mpinfo)
      call MPI_BCAST(cond%zshift,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      call MPI_BCAST(cond%NE,1,MPI_INTEGER,masterid,comm,mpinfo)
      call MPI_BCAST(cond%WriteEvectors,1,MPI_LOGICAL,masterid,comm,mpinfo)
      call MPI_BCAST(cond%CheckSymm,1,MPI_LOGICAL,masterid,comm,mpinfo) 
	  call MPI_BCAST(cond%Num_Ignore_KPoints,1,MPI_INTEGER,masterid,comm,mpinfo)
	  if (cond%Num_Ignore_KPoints > 0) then 
!		  if (.not. parallel%iammaster) allocate(cond%Ignore_KPoints(cond%Num_Ignore_KPoints))  
		  call MPI_BCAST(cond%Ignore_KPoints,cond%Num_Ignore_KPoints,MPI_INTEGER, masterid,comm,mpinfo)
	  endif 
      write(9,*) ' conductance(): After Bcasts.  Iammaster: ', parallel%iammaster 
        call myflush(9)
        call MPI_BARRIER(comm,mpinfo)
      write(9,*) ' conductance(): After barrier.  Iammaster: ', parallel%iammaster
	call myflush(9)

#endif

! Check that the number of eigenpairs to request from Arpack for the T(E) calculation 
! input from parsec.in (passed from usrinput()) is legal, and then broadcast it 
	if (parallel%iammaster) then
		write(9,*) ' conductance(): checking # epairs '
		call myflush(9)
		if (cond%NumEpairs > 0  .and.  cond%NumEpairs <= parallel%ndim) then
			cond%ProportionEpairs = real(cond%NumEpairs, dp) / real(parallel%ndim, dp)
		elseif (cond%ProportionEpairs > 0.0d0  .and. cond%ProportionEpairs <= 1.0d0) then 
			cond%NumEpairs = int( cond%ProportionEpairs * parallel%ndim )
		else 
			cond%ProportionEpairs = 0.05d0
			cond%NumEpairs = int( cond%ProportionEpairs * parallel%ndim )
			write(7,*) 'Conductance module: cond_epairs unspecified or invalid in parsec.in'
			write(7,*) 'Defaulting to 5% of total complex eigenpairs'
		endif 
	endif
#ifdef MPI
      write(9,*) ' After checking # epairs, before last Bcasts '
        call myflush(9)
        call MPI_BARRIER(comm,mpinfo)
      call MPI_BCAST(cond%NumEpairs,1,MPI_INTEGER,masterid,comm,mpinfo)
	call MPI_BCAST(cond%ProportionEpairs,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
	call MPI_BCAST(cond%shift,1,MPI_DOUBLE_PRECISION,masterid,comm,mpinfo)
      write(9,*) ' After # epairs, cond%shift Bcasts, before allocate cond%espectrum.  Cond%NE = ', cond%NE 
        call myflush(9)
        call MPI_BARRIER(comm,mpinfo)
#endif

	! Allocate the list of energies E at which T(E) will be calculated  
      allocate(cond%Espectrum(cond%NE),stat=alcstat)
      call alccheck('cond%Espectrum', cond%NE, alcstat)
	
	write(9,*) ' After allocate cond%espectrum, before Barrier '  
	call myflush(9) 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!     Start conductance calculation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!     

#ifdef MPI
      call MPI_BARRIER(comm,mpinfo)
#endif  

      if(parallel%iammaster) then
         write(7,*)
         write(7,*) "------------------------"
         write(7,*) "Conductance Calculation:"
         write(7,*) "------------------------"
         write(7,*)
         call myflush(7)
      endif

	call custom_date_time(datelabel,tBegin)
      
      write(9,*)
      write(9,*) "-----------------------------"
      write(9,*) "Begin Conductance Calculation"
      write(9,*) "-----------------------------"
      write(9,*)

      call myflush(9)
#ifdef MPI
      call MPI_BARRIER(comm,mpinfo)
#endif
write(9,*) ' conductance(): after Barrier '
call myflush(9)


!     The transport direction is the Z-direction.
      h = grid%step(3)

!!!!!!!!!!!!!!!!!!!!!!!!
!     Set up the diagonal of the complex absorbing potentials (CAPs) at the grid points
      
      lstdev2 = 2.0d0 * cond%gaussians(1)%stdev*cond%gaussians(1)%stdev
      rstdev2 = 2.0d0 * cond%gaussians(2)%stdev*cond%gaussians(2)%stdev

!     Set Z coordinate shift for CAPs  
         cond%zshift = clust%zatm(1) - cond%zshift  
		 cond%gaussians(1)%z0 = cond%gaussians(1)%z0 + cond%zshift
		 cond%gaussians(2)%z0 = cond%gaussians(2)%z0 + cond%zshift
		 if (abs(cond%zshift) > 0.35*h) then 
			write(7,*) 'Complex absorbing potential (CAP) centers z0 shifted in accordance with shift of atom coordinates!'  
			write(7,*) 'WARNING: This may cause CAPs to be outside the domain.  The new CAP centers are: '
			write(7,*) cond%gaussians(1)%z0, cond%gaussians(2)%z0 
		 endif 
      
      do igrid=1,parallel%mydim
         jgrid = igrid + parallel%irows(parallel%group_iam) - 1
!     NOTE: Gaussians are only in the transport or z-direction
         zz = (grid%shift(3) + grid%kz(jgrid))*h
		 
!     Set left CAP gammal
         r2 = zz - cond%gaussians(1)%z0
         r2 = r2 * r2
         cond%gammal(igrid) = cond%gaussians(1)%height*exp(-r2/lstdev2)
         
!     Set right CAP gammar
         r2 = zz - cond%gaussians(2)%z0
         r2 = r2 * r2
         cond%gammar(igrid) = cond%gaussians(2)%height*exp(-r2/rstdev2)
         
!     Total CAP gammafull 
         cond%gammafull(igrid) = cond%gammar(igrid) + cond%gammal(igrid)         
      enddo

      write(9,*) "Done making CAPs" 
      write(9,*) 'cond%writeEvectors = ', cond%writeEvectors
      call myflush(9)


	if (cond%writeEvectors) then 
! Gather CAPs into gammalall, gammarall for writing to disk 
! We do this so it's possible to combine partial eigensolutions from multiple TRANSEC runs, 
! as the grid ordering in the original runs may differ for different numbers of processors 
#ifdef MPI
      	call MPI_BARRIER(comm, mpinfo)

		! parallel%irows tells us the starting index (in 1-D arrays of grid points) of the grid points allocated to each processor 
		! MPI_GatherV takes as arguments the number of elements to be received from each process, and the displacement from the initial position of the receiving array 
		! (Because it's a displacement rather than a position, we pass parallel%irows(:) - 1) )
	
      	call MPI_GATHERV(cond%gammal,parallel%mydim,MPI_DOUBLE_PRECISION, &
      	gammalall,parallel%irows(1:parallel%group_size)-parallel%irows(0:parallel%group_size-1), &
      	parallel%irows(0:parallel%group_size-1)-1,MPI_DOUBLE_PRECISION,group_masterid,group_comm,mpinfo)

      	call MPI_GATHERV(cond%gammar,parallel%mydim,MPI_DOUBLE_PRECISION, &
      	gammarall,parallel%irows(1:parallel%group_size)-parallel%irows(0:parallel%group_size-1), &
      	parallel%irows(0:parallel%group_size-1)-1,MPI_DOUBLE_PRECISION,group_masterid,group_comm,mpinfo)

      	call MPI_BARRIER(comm, mpinfo)
#else
      	gammalall = cond%gammal
      	gammarall = cond%gammar
#endif
      	write(9,*) "Done gathering CAPs" 
      	call myflush(9)
		! Write gathered CAPs gammalall, gammarall to disk. 
		! We do this so it's possible to combine partial eigensolutions from multiple TRANSEC runs, 
		! as the grid ordering in the original runs may differ for different numbers of processors.   
		! (We perform this write before the loop over spin / k-points since the CAPs are indept of these.)  
		if (parallel%iammaster) then  
			open(88, file='CAP.dat', form='unformatted', status='replace')
			write(88) gammalall
			write(88) gammarall
			close(88)
		endif
	endif

write(9,*) ' Before auto-detecting symmetry'
call myflush(9)

! If current calculation is a periodic boundary condition (PBC) calculation set to make use 
! of R_z,pi (rotation by pi about z axis, (x,y,z)->(-x,-y,z), i.e. C_2 about the z axis) symmetry, 
! double-check that this symmetry is present  
! Note that in the future, this check may be extended to support additional symmetries (such as C_4 about the z axis)
if (elec_st%nkpt > 0 .and. parallel%iammaster .and. cond%CheckSymm) then  
     write(7,'(/,a)') 'Auto-detect symmetry properties for conductance calculation: calling atftmt()' 
 
	 ! Call atftmt() to check for R_z,pi symmetry.  Normally this routine is called in  
	 ! PARSEC via the driver routine symmetries() and the routine symgen(), which first  
	 ! determine the symmetries of the lattice vectors.  However, those routines do not  
	 ! search for symmetry in non-spherical cluster domains, or for rotational symmetry  
	 ! about a non-periodic axis like the transport axis z.  Here, we call atftmt() 
	 ! directly, while explicitly setting it to check for R_z,pi symmetry.   
	 ntrans = 1
	 ib(:) = 1
	 allocate(coorat(3, clust%type_num*maxval(clust%natmi)))
	 allocate(ity(clust%type_num*maxval(clust%natmi)))
	 
	 ! Reuse PARSEC's atftmt() routine to check for R_z,pi symmetry.  We call with the # of operations nc set to 1, 
	 ! such that the only operation r that is tested will correspond to R_z,pi.  The index ib is set on input to atftmt() 
	 ! such that ib(1) points to R_z,pi.
	 na = 0
	 do igrid = 1, clust%type_num
		ity((na+1):(na+clust%natmi(igrid))) = igrid
		na = na + clust%natmi(igrid)
	 enddo 
	 coorat(:,:) = zero
	 coorat(1,1:clust%atom_num) = clust%xatm(:)
	 coorat(2,1:clust%atom_num) = clust%yatm(:) 
	 coorat(3,1:clust%atom_num) = clust%zatm(:) 
	 r(:, :, :) = zero
	 r(1, 1, 1) = -one
	 r(1, 2, 2) = -one 
	 r(1, 3, 3) = +one  
	 call atftmt(ipr, symm%invlat, coorat, r, tnp, trans, ity, clust%atom_num, ib, 1, &
       -2, li, ntrans, invers_no, clust%type_num, maxval(clust%natmi))  
	 
	 ! If R_z,pi symmetry is absent, turn cond%RzSymm off  
	 if (ntrans >= 1) then   
		! In cases where we search for more than one symmetry, may need a more general criterion to judge whether R_z,pi symmetry is present 
		if (cond%RzSymm) then 
			write(7,*) 'R_z,pi (C_2) symmetry is present.'
		else 
			write(7,*) 'R_z,pi (C_2) symmetry found but will not be used because cond_use_inv_symm is off.'
		endif 
	 else
		write(7,*) 'WARNING: R_z,pi (C_2) symmetry not found.  Setting cond_use_inv_symm to false!'
		write(7,*) 'Transmission calculation for non-zero k-points will require roughly '
		write(7,*) 'double the computing time of the gamma point calculation.'
		cond%RzSymm = .false.  
	endif 
	write(7,*) 		
endif 
#ifdef MPI
! Broadcast updated value of cond%RzSymm to all processors
	write(9,*) ' Before barrier and Bcast of Rzsymm '
	call myflush(9)
      call MPI_BARRIER(comm,mpinfo)
	write(9,*) ' After barrier, before Bcast of Rzsym '
	call myflush(9)
      call MPI_BCAST(cond%RzSymm,1,MPI_LOGICAL,masterid,comm,mpinfo)
	write(9,*) ' After Bcast of Rzsym '
	call myflush(9)
#endif


      
!     Set up, and broadcast, the energy range for the T(E) calculation
      
      if(parallel%iammaster) then 
          
!     change bias from eV to Rydberg
            cond%Vbias = cond%Vbias / rydberg
      
!     change temperature from Kelvin (input in parsec.in) to Rydberg
            cond%Tcond = cond%Tcond * tempkry 
      
!     Elec_st%efermi contains the Fermi energy in Rydbergs
		cond%efermi = elec_st%efermi
 	      if (cond%NE < 2) then
	   	      cond%NE = 1
	            MuL = cond%efermi - 0.5d0 * (cond%Vbias)
        	      MuR = cond%efermi + 0.5d0 * (cond%Vbias)	
	      else
	            MuL = cond%efermi - 0.5d0 * DABS(cond%Vbias)
        	      MuR = cond%efermi + 0.5d0 * DABS(cond%Vbias)
	      endif
      
            TWidth = log(cond%FermCut - 1.0d0)*cond%Tcond;
            Emin = MuL - TWidth
            Emax = MuR + TWidth
      
	      if (cond%NE < 2) then
! Deal with the case where only one (or an illegal number of) energy point E has been 
! requested for the T(E) calculation in parsec.in
	   	      cond%NE = 1
		      dE = Emax - Emin  ! Value of dE irrelevant since only T(Emin) calculated
	      else
		      dE = (Emax - Emin) / (dble(cond%NE) - 1.0d0)
            endif

! Compute the spectrum of E points at which to do the T(E) calculation
            do i=1, cond%NE
                  cond%Espectrum(i) = Emin + dble((i-1))*dE
            enddo
      
      endif

! Broadcast the E spectrum
#ifdef MPI
write(9,*) ' Before Bcast of Espectrum '
call myflush(9)
      call MPI_BCAST(cond%Espectrum,cond%NE,MPI_DOUBLE_PRECISION, masterid,comm,mpinfo)
write(9,*) ' After Bcast of Espectrum '
call myflush(9)
      call MPI_BCAST(cond%Eigensolution_tol, 1, MPI_DOUBLE_PRECISION, masterid, comm, mpinfo)
      call MPI_BCAST(cond%efermi, 1, MPI_DOUBLE_PRECISION, masterid, comm, mpinfo)
write(9,*) ' After Bcast of Eigensolution_tol and EFermi '
call myflush(9)
      ! cond%Restart tells us which stage of the T(E) calculation to restart from (0 for no restart)
      call MPI_BCAST(cond%Restart, 1, MPI_INTEGER, masterid, comm, mpinfo) 
write(9,*) ' After Bcast of Restart, before allocates '
call myflush(9)
#endif





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Diagonalize H + i Gamma (condmatvec) !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      if(parallel%iammaster) then
            write(7,*) "----------------------------------------------------------------------"
            write(7,*) "Diagonalization of (H + i Gamma)"
            write(7,*) "----------------------------------------------------------------------"
            call myflush(7)
      endif

  nstate = cond%NumEpairs ! parallel%ndim * cond%ProportionEpairs

! Allocate memory for product GamLR of CAP matrices in diagonal basis of (H + i Gamma)
! Full CAP matrices Gamma -- to MPI_GATHER from r and w
  allocate(GamLR(nstate,nstate), stat=alcstat) 
  call alccheck('gamlr', nstate**2, alcstat)

! Allocate memory for eigenvalues of (H + i Gamma)
  allocate(eval(nstate), stat=alcstat) 
  call alccheck('eval', nstate, alcstat)
! If no R_z,pi (rotation by pi about z axis, (x,y,z)->(-x,-y,z)) symmetry, we solve for -k also  
! Note that every k-point except k=0 requires this -- a potential load-balancing issue 
  if ( (elec_st%nkpt > 0) .and. elec_st%cplx .and. (.not. cond%RzSymm) ) then 
  	allocate(evalNoRz(nstate), stat=alcstat) 
  	call alccheck('evalNoRz', nstate, alcstat)
	allocate(bestMatchInd(nstate), stat=alcstat)
	call alccheck('bestMatchInd', nstate, alcstat)
  endif 

write(9,*) ' After allocates '
call myflush(9) 
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!! Begin loops over spins and k-points
iEigensolution = 0
SpinLoop: do isp = 1, elec_st%nspin/elec_st%mxwd 
  write(9,*) ' isp ', isp 
  call myflush(9)

  KPointLoop: do kplp = 1, max(elec_st%nkpt,1) 
! Note that a loop over symmetry representations is unnecessary because ignoresym is set to true when doing conductance 
! calculations, so elec_st%nrep should be 1   

	iEigensolution = iEigensolution + 1
	
	! Determine kpt_PBC and SignK 
	! If this is a periodic calculation, kpt_PBC is a flag signifying the current k-point is non-zero, i.e. not the Gamma point.  
	! If R_z,pi symmetry is also absent, SignK denotes which sign(s) of the current k-point should be performed on this MPI group.   
	kpt_PBC = (elec_st%cplx) .and. (elec_st%nkpt > 0) 
	if (kpt_PBC) then 
		! Check whether current k-point is the Gamma point.  The elec_st%kpts array is only allocated 
		! for a periodic calculation, so to avoid the possibility of seg faults, only attempt to access 
		! elec_st%kpts after confirming calculation is periodic.  
		kpt_PBC = kpt_PBC .and. any(abs(elec_st%kpts(:,kplp)) > 5.d-5) 
  
		! Determine SignK 
		if (cond%RzSymm .or. (.not. kpt_PBC)) then 
			! If R_z,pi symmetry is present, only perform eigensolution for +k, since the -k eigenpairs 
			! are related by symmetry.  For the Gamma point, -k = +k = 0.  
			SignK = SYMM_K 
		else 
			! If R_z,pi symmetry is absent, the default is to perform both +k and -k eigensolutions in serial  
			! on the same MPI group.  However, the exception is when parallelizing them improves the overall
			! load-balancing of the conductance calculation.  This is signified by cond%Parallelize_Over_SignK > 0.  
			! In this case, the first cond%Parallelize_Over_SignK eigensolutions are performed in parallel with each other.  
			if (iEigensolution > cond%Parallelize_Over_SignK) then 
				SignK = BOTH_K  ! Current group is assigned to both +k and -k, in serial  
			elseif (elec_st%eig(1,kplp,isp)%group == parallel%mygroup) then 
				SignK = POSITIVE_K  ! Current group is assigned to +k 
			elseif (elec_st%eig(1,kplp,isp)%group + 1 == parallel%mygroup) then 
				SignK = NEGATIVE_K  ! Current group is assigned to -k 
			else 
				SignK = NO_K  ! Current group is not assigned to this k-point  
			endif 
			
			iEigensolution = iEigensolution + 1
		endif 
	endif 

	write(9,*) ' kplp ', kplp
	call myflush(9)

	write(9,*) ' elec_st%eig(1,kplp,isp)%group ', elec_st%eig(1,kplp,isp)%group
	call myflush(9)
	write(9,*) ' parallel%mygroup ', parallel%mygroup 
	call myflush(9)
	write(9,*) ' cond%Num_Ignore_KPoints ', cond%Num_Ignore_KPoints
	call myflush(9) 
	write(9,*) ' cond%Ignore_KPoints(1:cond%Num_Ignore_KPoints) ', cond%Ignore_KPoints(1:cond%Num_Ignore_KPoints)
	call myflush(9)

! As in eigval(), a serial job would iterate over kplp, isp serially, but a parallel 
! job should only work if the current kplp, isp are assigned to current MPI group.  
! IS IT POSSIBLE / ADVISABLE TO HAVE elec_st%eig%group BE DIFFERENT ON THE GROUPS DOING +K AND -K ??  
  if ( (elec_st%eig(1,kplp,isp)%group /= parallel%mygroup .and. SignK /= NEGATIVE_K) .or. &
	(cond%Num_Ignore_KPoints > 0 .and. any(cond%Ignore_KPoints(1:cond%Num_Ignore_KPoints) == kplp)) ) cycle


	write(9,*) ' About to determine spinID string '
	call myflush(9)

  maxmvs = 135000

  if ( isp == 1 .and. elec_st%nspin > elec_st%mxwd) then
	spinID = 'UP' 
  elseif (isp == 2) then
  	spinID = 'DN'
  else
  	spinID = 'TOT' 
  endif 
  if (elec_st%nkpt > 0) write(spinID, "(a,a,i0)") trim(spinID), '.Kpt', kplp
  write(9,*) ' Diagonalization of (H + i Gamma): ', trim(spinID)
  call myflush(9)


  nec = nstate
  ierr = 0

  GamLR(:,:) = zzero
  resid_restart = zzero


! As in eigval(), ensure the spin-appropriate Hartree potential is in KS Hamiltonian 
  if (parallel%mxwd == 1) then
	if (elec_st%nkpt == 0) then
      	call dcopy(parallel%mydim,pot%vnew(1,isp),1,solver%adiag,1)
            solver%adiag(:) = solver%adiag(:) + sum(solver%coe2(0,:))
      else
      	call dcopy(parallel%mydim,pot%vnew(1,isp),1,solver%adiag,1)
            solver%adiag(:) = solver%adiag(:) + sum(solver%coe2(0,:)) + &
            	elec_st%kpts(1,kplp)*elec_st%kpts(1,kplp) + &
                  elec_st%kpts(2,kplp)*elec_st%kpts(2,kplp) + &
                  elec_st%kpts(3,kplp)*elec_st%kpts(3,kplp)
      endif
  endif

  write(9,*) ' After setting adiag = ', solver%adiag(1:8) 

Restart_If: if (cond%Restart < 1) then
! Allocate memory for eigenvectors of (H + i Gamma)
  if (SignK == BOTH_K .or. SignK == POSITIVE_K .or. SignK == SYMM_K) then 
	  allocate(evec(parallel%ldn,nstate), stat=alcstat)
	  call alccheck('evec', parallel%ldn*nstate, alcstat)
  endif 
  !!!!! If no R_z,pi (rotation by pi about z axis, (x,y,z)->(-x,-y,z)) symmetry, solve for -k also  
  !!!! Note that every k-point except k=0 requires this -- a potential load-balancing issue 
  if ( (elec_st%nkpt > 0) .and. elec_st%cplx .and. (.not. cond%RzSymm) ) then !.and. (SignK == BOTH_K .or. SignK == NEGATIVE_K) ) then 
  	allocate(evecNoRz(parallel%ldn,nstate), stat=alcstat)
  	call alccheck('evecNoRz', parallel%ldn*nstate, alcstat)
  endif  

  if (SignK == BOTH_K .or. SignK == POSITIVE_K .or. SignK == SYMM_K) then 
	if (parallel%iamgmaster) then
		 write(7,*) 'Spin '//trim(spinID)//': Arpack begin - Call zarpk_gamma_diag()'
		 call myflush(7)
		 call custom_date_time(datelabel,tBeginK)
	endif

	! Solve for eigenvalues & vectors of (H + i Gamma)
	! Call Arpack.  Should modify to use whichever (same) solver PARSEC does in eigval
	  call zarpk_gamma_diag( kplp, elec_st%cplx, +1, parallel, nstate, nec, cond, maxmvs,eval,evec, & 
					 cond%Eigensolution_tol,resid_restart,ipr,ierr  )

#ifdef MPI
	call MPI_BARRIER(group_comm, mpinfo)
#endif

	  write(9,*) 'nec= ', nec
  endif 


!!!!! If no R_z,pi (rotation by pi about z axis, (x,y,z)->(-x,-y,z)) symmetry, solve for -k also 
! For parallel computations, we use the same MPI group to solve for both +k and -k.   
! Note this could cause a load-balancing, and memory-balancing, issue, since the group solving 
! for the gamma point has no -k to solve for.  But all in all, this would seem a simpler 
! strategy than trying to set up a new set of MPI groups different from those used by PARSEC
! to do the DFT, that may divide differently into the # of processors.  (Because the DFT 
! problem has no CAPs, PARSEC can rely on time-reversal symmetry for -k).  Moreover, 
! T_k(E) depends on both +k and -k, so we avoid inter-group communication by solving
! both of these in the same group.  
if (kpt_PBC .and. (.not. cond%RzSymm) ) then 
	if (SignK == BOTH_K .or. SignK == NEGATIVE_K) then   
		if (parallel%iamgmaster) then
			write(7,*) 'Spin '//trim(spinID)//': Solve for -k points - Call zarpk_gamma_diag()'
			call myflush(7)
			call custom_date_time(datelabel,tBeginK)
		endif

		  maxmvs = 135000
		  nec = nstate

		  call zarpk_gamma_diag( kplp, elec_st%cplx, -1, parallel, nstate, nec, cond, maxmvs,evalNoRz,evecNoRz, & 
						 cond%Eigensolution_tol,resid_restart,ipr,ierr  ) 
		!!! NOTE: At present, we use the RetAdv flag to perform eigensolution for -k.   
		! Strictly speaking, this is a misnomer.  The advanced Green's function should be just 
		! G^a = G* = G^dagger, where G is the retarded Green's function, since we work with 
		! H_k |k><k| + H_{-k} |-k><-k|, which is real-symmetric before the addition of CAPs.  
		! But condmatvec just applies H_k + i*Gamma, so we use RetAdv to get H_k - i*Gamma and then 
		! can complex-conjugate this to get H_{-k} + i*Gamma.  

#ifdef MPI
		  call MPI_BARRIER(group_comm, mpinfo)
#endif

		  write(9,*) 'nec= ', nec
	endif 	  

	! In the case of parallelization over the sign of k, communicate the -k 
	! eigenpairs to the group doing +k.  Subsequently, we will skip the rest of the work 
	! for this k-point if SignK == NEGATIVE_K.  
	if (SignK == NEGATIVE_K) then 
#ifdef MPI
		! Should check irows of destination PE.  Shouldn't really use irows as tag 
		call MPI_SEND(evalNoRz, nstate, MPI_DOUBLE_COMPLEX, &
			parallel%gmap(parallel%group_iam+1, parallel%mygroup), &
			parallel%irows(parallel%group_iam), comm, mpinfo) 
		call MPI_SEND(evecNoRz, parallel%ldn*nstate, MPI_DOUBLE_COMPLEX, &
			parallel%gmap(parallel%group_iam+1, parallel%mygroup), &
			parallel%irows(parallel%group_iam), comm, mpinfo)
#endif
	else 
		if (SignK == POSITIVE_K) then 
#ifdef MPI
		! Should check irows of destination PE.  Shouldn't really use irows as tag 
		call MPI_RECV(evalNoRz, nstate, MPI_DOUBLE_COMPLEX, &
			parallel%gmap(parallel%group_iam+1, parallel%mygroup+2), &
			parallel%irows(parallel%group_iam), comm, mpistatus, mpinfo)
		call MPI_RECV(evecNoRz, parallel%ldn*nstate, MPI_DOUBLE_COMPLEX, &
			parallel%gmap(parallel%group_iam+1, parallel%mygroup+2), &
			parallel%irows(parallel%group_iam), comm, mpistatus, mpinfo) 
#endif
		endif 
		if (parallel%iamgmaster) then
			write(7,*) 'Spin '//trim(spinID)// &
			': Align solved eigenpairs for -k points with +k points - call align_epair_kpt()'
			call myflush(7)
		endif
		call align_epair_kpt( nstate, eval, evalNoRz, parallel, bestMatchInd )   
	endif 
endif



if (parallel%iamgmaster) then
     call custom_date_time(datelabel,tEnd)
     write(7,*) 'Spin '//trim(spinID)//': Arpack finished - Current date/time: ',datelabel,' UTC'
     write(7,*) 'Arpack time: ', tEnd - tBeginK, 's '
     call myflush(7)

     fileUnit = 88 + parallel%mygroup
     open(fileUnit, file='Eval.'//trim(spinID)//'.dat', form='unformatted', status='replace')
     write(fileUnit) eval
     close(fileUnit)
endif

! MODIFY SPINID TO SPECIFY SIGNK ?  
if (parallel%iamgmaster) then
     write(7,*) 'Spin '//trim(spinID)// & 
     ': Orthonormalize solved eigenvectors by bi-orthogonal inner product - call orthonormalize_GramSchmidt_writeEvecs()'
     call myflush(7)
endif

SignK_If: if (SignK == BOTH_K .or. SignK == POSITIVE_K .or. SignK == SYMM_K) then   
  if (cond%writeEvectors .and. kpt_PBC .and. (.not. cond%RzSymm) ) then 
	call orthonormalize_GramSchmidt_writeEvecs(evec, eval, nec, cond, parallel, grid, kpt_PBC, &
		kplp, elec_st%cplx, cond%gammafull, & 
#ifdef MPI
		mpinfo, &
#endif
		spinID, evecNoRz, evalNoRz, bestMatchInd)
  elseif(cond%writeEvectors) then 
	call orthonormalize_GramSchmidt_writeEvecs(evec, eval, nec, cond, parallel, grid, kpt_PBC, &
		kplp, elec_st%cplx, cond%gammafull, & 
#ifdef MPI
		mpinfo, &
#endif
		spinID)
  elseif(kpt_PBC .and. (.not. cond%RzSymm)) then 
	call orthonormalize_GramSchmidt_writeEvecs(evec, eval, nec, cond, parallel, grid, kpt_PBC, &
		kplp, elec_st%cplx, cond%gammafull, & 
#ifdef MPI
		mpinfo, &
#endif
		evecNoRz=evecNoRz, evalNoRz=evalNoRz, bestMatchInd=bestMatchInd)
  else
#ifdef MPI
	call orthonormalize_GramSchmidt_writeEvecs(evec, eval, nec, cond, parallel, grid, kpt_PBC, &
		kplp, elec_st%cplx, cond%gammafull, mpinfo)
#else
	call orthonormalize_GramSchmidt_writeEvecs(evec, eval, nec, cond, parallel, grid, kpt_PBC, &
		kplp, elec_st%cplx, cond%gammafull)
#endif
 endif 
 
if (parallel%iamgmaster) then
     write(7,*) 'Spin '//trim(spinID)//': Convert CAPs to diagonal basis of (H + i Gamma) - call basis_change_CAPs()'
     call myflush(7)
endif
#ifdef MPI
  call basis_change_CAPs(nstate, parallel, cond, evec, & 
  	(kpt_PBC .and. (.not. cond%Rzsymm)), evecNoRz, bestMatchInd, GamLR, mpinfo)
#else
  call basis_change_CAPs(nstate, parallel, cond, evec, &
  	(kpt_PBC .and. (.not. cond%Rzsymm)), evecNoRz, bestMatchInd, GamLR)
#endif
endif SignK_If 


! Deallocate eigenvectors of (H + i Gamma) -- no longer needed
  if (allocated (evec))  deallocate(evec)
! If no R_z,pi (rotation by pi about z axis, (x,y,z)->(-x,-y,z)) symmetry, deallocate for -k also  
  if (allocated (evecNoRz))  deallocate(evecNoRz)

  if (parallel%iamgmaster) then

    open(fileUnit, file='GammaLR.'//trim(spinID)//'.dat', form='unformatted', status='replace')

! Write GamLR to disk
    write(fileUnit) GamLR
    close(fileUnit)
  endif


else  
  ! If Cond%Restart > 0, read in GammaL, GammaR, eigenvalues
  if ( (SignK == BOTH_K .or. SignK == POSITIVE_K .or. SignK == SYMM_K) &
	.and. parallel%iamgmaster) then
    open(fileUnit, file='GammaLR.'//trim(spinID)//'.dat', form='unformatted', status='old')
    read(fileUnit) GamLR
    close(fileUnit)

    open(fileUnit, file='Eval.'//trim(spinID)//'.dat', form='unformatted', status='old')
    read(fileUnit) eval
    close(fileUnit) 
  endif

endif Restart_If ! Cond%Restart




!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Compute Transmission TE
if (parallel%iamgmaster) then
	 ! Output update: beginning transmission spectrum computation 
	 if (parallel%iammaster) then
		! receive spinID 
		write(7,*) "----------------------------------------------------------------------"
		write(7,*) 'Spin '//trim(spinID)//': Compute Transmission spectrum - call compute_TE()' 
		call myflush(7)
	 else 
		! Send spinID to master 
		write(9,*) "----------------------------------------------------------------------"
		write(9,*) 'Spin '//trim(spinID)//': Compute Transmission spectrum - call compute_TE()' 
		call myflush(9)
	 endif 

     fileUnit = 1001 + parallel%mygroup
     call compute_TE(nstate, GamLR, cond, eval, spinID, fileUnit, ipr)

	 ! Output update: transmission spectrum computation completed  
	 if (parallel%iammaster) then
		! receive spinID 
		write(7,*) 'Spin '//trim(spinID)//': Transmission spectrum computation complete' 
		write(7,*) "----------------------------------------------------------------------"
		call myflush(7)
	 else 
		! Send spinID to master 
		write(9,*) 'Spin '//trim(spinID)//': Transmission spectrum computation complete' 
		write(9,*) "----------------------------------------------------------------------"
		call myflush(9)
	 endif 
endif 

#ifdef MPI
call MPI_BARRIER(group_comm, mpinfo)
#endif

if (parallel%iamgmaster) then
     call custom_date_time(datelabel,tEnd)
     write(7,*) 'Spin '//trim(spinID)//': TRANSEC finished' 
     call myflush(7)
endif
write(9,*) 'TRANSEC finished'
call myflush(9)


enddo KPointLoop
enddo SpinLoop



#ifdef MPI
call MPI_BARRIER(comm, mpinfo)
#endif

if (parallel%iammaster) then
     call custom_date_time(datelabel,tEnd)
     write(7,*) 'TRANSEC finished - Current date/time: ',datelabel,' UTC'
     write(7,*) 'Total TRANSEC time: ', tEnd - tBegin, 's '
     call myflush(7)
endif



!!!!!!!!!!!!!!!!!!!!!!!!!!!!        
!     Memory deallocations
!     
      call destroy_conductance_param (cond)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!         
!     Deallocate local arrays
!     
      if (allocated (GamLR)) deallocate(GamLR)
      if (allocated (eval))  deallocate(eval)
      if (allocated (evalNoRz))  deallocate(evalNoRz)
	  if (allocated (bestMatchInd))  deallocate(bestMatchInd)
	  
!!!!!!!!!!!!!!!!!!!!!!!!!!!
!     NEED TO CONVERT POTENTIAL PARAMETERS TO ATOMIC UNITS?
!     
end subroutine postSCF_conductance







subroutine init_random_seed()

            INTEGER :: i, n, clock
            INTEGER, DIMENSION(:), ALLOCATABLE :: seed
          
            CALL RANDOM_SEED(size = n)
            ALLOCATE(seed(n))
          
            CALL SYSTEM_CLOCK(COUNT=clock)
          
            seed = clock + 37 * (/ (i - 1, i = 1, n) /)
            CALL RANDOM_SEED(PUT = seed)
          
            DEALLOCATE(seed)

end subroutine init_random_seed



#ifdef MPI 
subroutine orthonormalize_GramSchmidt_writeEvecs(evec, eval, nstate, cond, parallel, grid, kpt_PBC, kplp, Hks_cplx, &
	gammafull, mpinfo, spinID, evecNoRz, evalNoRz, bestMatchInd)
#else
subroutine orthonormalize_GramSchmidt_writeEvecs(evec, eval, nstate, cond, parallel, grid, kpt_PBC, kplp, Hks_cplx, &
	gammafull, spinID, evecNoRz, evalNoRz, bestMatchInd) 
#endif
! Check whether Gram-Schmidt is needed to orthogonalize eigenvectors in same eigenspace (based on q^t dot q').  
! Also normalize eigenvectors based on q^t dot q = 1.  

	use constants
	use conductance_module
	use parallel_data_module
	use grid_module

#ifdef MPI
      use mpi 
#endif

! Number of solved-for eigenpairs, and size of evec, eval arrays
	integer, intent(in) :: nstate 
! Parameters defining the conductance calculation to perform 
      type(conductance_param), intent(in) :: cond
! The parallel environment
	type(parallel_data), intent(in) :: parallel
! The grid points and mappings
	type(grid_data), intent(in) :: grid
	integer, intent(in) :: kplp 
	! kpt_PBC is a flag for the current k-pt being non-zero, i.e. not the Gamma point.  
	! Hks_cplx is a flag for complex wavefunctions.  It must be true if kpt_PBC is true.  
	logical, intent(in) :: kpt_PBC, Hks_cplx 
	real(dp), intent(in)  :: gammafull(parallel%ldn)
! For storing computed eigenvalues & eigenvectors of H + i Gamma
	complex(dpc), intent(in) :: eval(nstate)
	complex(dpc), intent(inout) :: evec(parallel%ldn, nstate)
	complex(dpc), intent(in), optional :: evalNoRz(nstate)
	complex(dpc), intent(inout), optional :: evecNoRz(parallel%ldn, nstate) 
#ifdef MPI
	integer, intent(inout) :: mpinfo   
#endif
! Spin identifier for filenames, output
	character(len=25), optional, intent(in) :: spinID 
	integer, intent(in), optional :: bestMatchInd(nstate)
	
	! Work variables to hold normalizations and dot products 
	complex(dpc) :: tau, tau2  
    ! The full grid of the ith eigenvector (gathered from all processors) for writing to disk  
	complex(dpc), dimension(parallel%ndim) :: evecall
    ! The full grid of previous eigenvectors (gathered from all processors) for writing to disk  
	complex(dpc), allocatable, dimension(:,:) :: evecallOld
	complex(dpc) :: A(parallel%ldn)
	! A version of grid%indexg() that actually is an inverse of grid%kx, ky, and kz  
	integer :: kInverse(grid%nxmin:grid%nxmax, grid%nymin:grid%nymax, grid%nzmin:grid%nzmax)    
	integer :: i, j, alcstat  
	integer, parameter :: NumOld = 700 ! 25 
	! File unit to be used for writing eigenvectors    
	integer :: fileUnit

!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
!     Interface for routine with optional parameters 
	interface 
#ifdef MPI
		subroutine normalize_kpt(evecall, grid, kInverse, tau, parallel, mpinfo, evecall2) 
#else
		subroutine normalize_kpt(evecall, grid, kInverse, tau, parallel, evecall2) 
#endif 
			use constants
			use parallel_data_module
			use grid_module
			implicit none
			type(grid_data), intent(in) :: grid
			integer, intent(in) :: kInverse(grid%nxmin:grid%nxmax, grid%nymin:grid%nymax, grid%nzmin:grid%nzmax)    
			type(parallel_data), intent(in) :: parallel
			complex(dpc), intent(in) :: evecall(parallel%ndim)
			complex(dpc), intent(in), optional :: evecall2(parallel%ndim)
			complex(dpc), intent(out) :: tau
#ifdef MPI
			integer, intent(inout) :: mpinfo   
#endif
		end subroutine
	end interface
	
	if (kpt_PBC .and. cond%RzSymm) then ! kplp > 1) then
	! Set up kInverse, a version of grid%indexg() that actually is an inverse of grid%kx, ky, and kz  
		do i = 1, parallel%ndim
			kInverse(grid%kx(i), grid%ky(i), grid%kz(i)) = i  
		enddo 
	endif 

	if (parallel%iamgmaster) then
		if (kpt_PBC .and. cond%RzSymm) then 
			allocate(evecallOld(parallel%ndim, NumOld), stat=alcstat)
			call alccheck('evecallOld',parallel%ndim*NumOld,alcstat)
			evecallOld(:,:) = zzero
		endif 
		if (cond%writeEvectors) open(fileUnit, file='Evec.'//trim(spinID)//'.dat', form='unformatted', status='replace')
	endif 
		
	! EvecLoop is the main loop in this routine, to gather, orthogonalize, and normalize the ith eigenvector
	EvecLoop: do i = 1, nstate

	! Gather the eigenvectors in order to write to disk and/or for orthonormalization with k-points.   
	! We write to disk so it's possible to combine partial eigensolutions from multiple TRANSEC runs.  
	! Writing the eigenvectors is included in this orthonormalization subroutine in order to minimize 
	! the total number of MPI_gather's, and because the eigenvectors have now been fully prepared.   
		if (cond%WriteEvectors .or. kpt_PBC) then ! kplp > 1)) then
#ifdef MPI
			call MPI_GATHERV(evec(:,i), parallel%mydim, MPI_DOUBLE_COMPLEX, &
				evecall,parallel%irows(1:parallel%group_size)-parallel%irows(0:parallel%group_size-1), &
				parallel%irows(0:parallel%group_size-1)-1,MPI_DOUBLE_COMPLEX,parallel%group_master,parallel%group_comm,mpinfo)
#else
			evecall = evec(:, i) 
#endif		

		endif 

		! OrthogonalizationLoop checks whether vector i is in same eigenspace as any previous 
		! vectors j, and if so, orthogonalizes via Gram-Schmidt.    
		OrthogonalizationLoop: do j = i-1, 1, -1
		! The eigenvalues are sorted, so we exit Gram-Schmidt loop as soon as the jth eval 
		! differs from the ith by more than a tolerance.  
		! Consider broadcasting eval array before i, j loops to guarantee all cores 
		! agree on this test result.  (In practice this is unnecessary, since the evals 
		! received from ARPACK are identical on each core).   
			if ( abs(eval(i)-eval(j)) > 1.d-2 )  exit  

			! Orthogonalize (based on q^t dot q')
			if (.not. kpt_PBC) then 
			! Case of no PBCs / gamma point with PBCs  
				tau = dot_product(conjg(evec(1:parallel%mydim,j)), evec(1:parallel%mydim,i))
#ifdef MPI
				call zpsum(tau, 1, parallel%group_size, parallel%group_comm)
				call MPI_BARRIER(parallel%group_comm, mpinfo)  
#endif
				write(9,*) ' orthonormalize_GramSchmidt_writeEvecs(): ', i, j, ', dot product = ', tau   
!				tau = dot_product(conjg(evec(1:parallel%mydim,j-2)), evec(1:parallel%mydim,i))
#ifdef MPI
!				call zpsum(tau, 1, parallel%group_size, parallel%group_comm)
!				call MPI_BARRIER(parallel%group_comm, mpinfo)
#endif
!				write(9,*) ' orthonormalize_GramSchmidt_writeEvecs():    for comparison: ', i, j-2, ', dot product = ', tau   
			elseif(cond%RzSymm) then 
			! Case of PBCs with R_z,pi symmetry 
				if ((i-j) <= NumOld) then 
				! If i <= NumOld + 1, at this point within EvecLoop, evecallOld only holds (i-1) old eigenvectors.  
				! But the bounds of OrthogonalizationLoop guarantee that 1 <= i-j <= i-1, so evecallOld(:, i-j) is well-defined.  
				! If i > NumOld + 1, it's possible that evecallOld(:, i-j) doesn't exist.  But NumOld has been set large enough 
				! that we're unlikely to find a NumOld-fold degeneracy, so OrthogonalizationLoop will very likely terminate 
				! before i-j reaches NumOld.  
#ifdef MPI
					call normalize_kpt(evecall, grid, kInverse, tau, parallel, mpinfo, evecallOld(1:parallel%ndim, i-j))
#else
					call normalize_kpt(evecall, grid, kInverse, tau, parallel, evecallOld(1:parallel%ndim, i-j))
#endif		
				else 
					tau = zzero 
				endif 
			else 
			! Case of PBCs without R_z,pi symmetry   
				call normalize_kpt_NoRz(evec(:,i), evecNoRz(:,bestMatchInd(i-j)), tau, parallel)  
			endif 
			
			evec(:,i) = evec(:,i) - tau * evec(:,j)
			if (kpt_PBC  .and.  cond%RzSymm  .and.  parallel%iamgmaster  .and.  (i-j) <= NumOld) &
				evecall(1:parallel%ndim) = evecall(1:parallel%ndim) - tau * evecallOld(1:parallel%ndim, i-j)

		enddo OrthogonalizationLoop

	! Normalize ith eigenvector (based on q^t dot q = 1)
		if (.not. kpt_PBC) then 
		! Case of no PBCs / gamma point in PBC calculation 
			tau = dot_product(conjg(evec(1:parallel%mydim,i)), evec(1:parallel%mydim,i))
			call zpsum(tau, 1, parallel%group_size, parallel%group_comm)
			write(9,*) ' orthonormalize_GramSchmidt_writeEvecs():    ', i, 'dotted with itself = ', tau   
		else  
#ifdef DEBUG  
			write(9,*) 'Call normalize_kpt()'
			call myflush(9)
#endif 
			if ( cond%RzSymm ) then 
			! For cases with PBCs and R_z,pi rotational symmetry  
#ifdef MPI
				call normalize_kpt(evecall, grid, kInverse, tau, parallel, mpinfo)
#else
				call normalize_kpt(evecall, grid, kInverse, tau, parallel)
#endif
			else 
			! For cases with PBCs where V_KS lacks R_z,pi rotational symmetry  
				if (bestMatchInd(i) < 0) then 
					tau = zone
				else
					call normalize_kpt_NoRz(evec(:,i), evecNoRz(:,bestMatchInd(i)), tau, parallel)
					evecNoRz(:, bestMatchInd(i)) = evecNoRz(:, bestMatchInd(i)) / sqrt(tau)
				endif 
			endif 
		endif 
		
		evec(:, i) = evec(:, i) / sqrt(tau) 
		evecall = evecall / sqrt(tau)  
  
	! Verify that q is an eigenvector with eigenvalue epsilon
	! Compute residual tau = | (H - i*Gamma - epsilon) q |  
		A(:) = zzero 
		if (Hks_cplx .or. kpt_PBC) then
			call zcondmatvec(kplp, gammafull, 0.d0, evec(1:parallel%mydim,i), A, +1, parallel)
		else
			call condmatvec(kplp, gammafull, 0.d0, evec(1:parallel%mydim,i), A, +1, parallel)
		endif 
		A(:) = A(:) + (eval(i) * evec(:,i))
		tau = zzero  
		tau = dot_product(conjg(A(1:parallel%mydim)), A(1:parallel%mydim))  
		call zpsum(tau, 1, parallel%group_size, parallel%group_comm)
		tau2 = zzero  
		if (kpt_PBC .and. (.not. cond%Rzsymm) .and. present(evecNoRz) ) then 
			A(:) = zzero 
			call zcondmatvec(kplp, gammafull, 0.d0, conjg(evecNoRz(1:parallel%mydim,i)), A, -1, parallel)
			if (present(evalNoRz)) then
				A(:) = conjg(A(:)) + (evalNoRz(i) * evecNoRz(:,i))
			else
				A(:) = conjg(A(:)) + (eval(i) * evecNoRz(:,i))
			endif 
			tau2 = dot_product(conjg(A(1:parallel%mydim)), A(1:parallel%mydim))  
			call zpsum(tau2, 1, parallel%group_size, parallel%group_comm)
		endif 
#ifdef MPI 
		call MPI_BARRIER(parallel%group_comm, mpinfo)
#endif 
	! Output the residual tau 
		if (parallel%iamgmaster) then 
#ifdef DEBUG
			write(7,*) '  orthonormalize_GramSchmidt_writeEvecs(): |Residual| ', i, ' = ', abs(tau)  
			if (kpt_PBC .and. (.not. cond%Rzsymm) ) write(7,*) '    For -k point: |Residual| ', &
				i, ' = ', abs(tau2)  
#endif 	
	! If q is not an eigenvector, output a warning  
			if (abs(tau) > 5.d-6) write(7,*) '  orthonormalize_GramSchmidt_writeEvecs(): WARNING: Residual ', &
				i, ' = ', tau, ' IS LARGE, POSSIBLE SPURIOUS EIGENPAIR. '   
			if (kpt_PBC .and. (.not. cond%Rzsymm) .and. abs(tau2) > 5.d-6) &
				write(7,*) '  orthonormalize_GramSchmidt_writeEvecs(): WARNING: Residual ', &
				i, ' = ', tau2, ' for -k point IS LARGE, POSSIBLE SPURIOUS EIGENPAIR. '   
		endif 

		! Write curent eigenvector to disk so it's possible to combine partial eigensolutions from 
		! multiple TRANSEC runs.  Writing the eigenvectors is included in this subroutine in order to 
		! minimize the number of MPI_gather's, and because the eigenvectors have now been fully processed.   
		if (cond%writeEvectors .and. parallel%iamgmaster) write(fileUnit) evecall
				
		! For cases with PBCs and R_z,pi symmetry, keep track of a number of old gathered eigenvectors 
		! for use with orthogonalization (i.e., in OrthogonalizationLoop above) 
		if (kpt_PBC .and. cond%RzSymm .and. parallel%iamgmaster) then 
			evecallOld(1:parallel%ndim, 2:NumOld) = &
				evecallOld(1:parallel%ndim, 1:(NumOld-1))
			evecallOld(1:parallel%ndim, 1) = evecall(1:parallel%ndim) 
		endif 

	enddo EvecLoop 
		
	if (parallel%iamgmaster) then 
		if (kpt_PBC .and. cond%RzSymm .and. allocated(evecallOld)) deallocate(evecallOld)
		if (cond%writeEvectors) close(fileUnit)
	endif 
  
end subroutine orthonormalize_GramSchmidt_writeEvecs



subroutine normalize_kpt_NoRz(evec_i, evecNoRz_i, tau, parallel) 

! This subroutine is for cases where u_{-k} != R_z,pi u_k, where R_z,pi is 
! rotation by pi about the z-axis, i.e. for cases where V_KS lacks R_z,pi symmetry.   
! In these cases we first need to solve explicitly for u_{-k} and then use it in this routine.  
  use constants 
  use parallel_data_module
#ifdef MPI
      use mpi 
#endif

  implicit none

! The parallel environment
  type(parallel_data), intent(in) :: parallel
! Eigenvectors of H + i Gamma
  complex(dpc), intent(in) :: evec_i(parallel%ldn), evecNoRz_i(parallel%ldn)
! Computed normalization constant 
  complex(dpc), intent(out) :: tau

     ! Normalize (based on q^t dot q = 1)
     tau = dot_product(conjg(evec_i(1:parallel%mydim)), evecNoRz_i(1:parallel%mydim))
#ifdef MPI
     call zpsum(tau, 1, parallel%group_size, parallel%group_comm)
#endif

#ifdef DEBUG  
     ! Write debug info 
	 write(9,*) 'normalize_kpt_NoRz(): tau = ', tau 
     call myflush(9)
#endif
	 
end subroutine normalize_kpt_NoRz


#ifdef MPI
subroutine normalize_kpt(evecall, grid, kInverse, tau, parallel, mpinfo, evecall2) 
#else
subroutine normalize_kpt(evecall, grid, kInverse, tau, parallel, evecall2) 
#endif 
! Normalize the periodic part of the Bloch function, which is an eigenvector 
! u_n,k of the k-point-restricted Hamiltonian H_k.   
! We normalize based on u dot u (i.e. no complex conjugation).  But we assume that 
! linear combinations v_k+ = u_k e^ikx + u_{-k} e^-ikx and 
! v_k- = (u_k e^ikx - u_{-k} e^-ikx)/i are taken, such that v_k is real 
! in the limit that the CAPs vanish.  Because of the Bloch e^ikx factors, the dot 
! product normalization couples u_k to u_{-k}, rather than to itself.  

! IMPORTANT NOTE:  Presently this subroutine assumes that u_{-k} = R_z,pi u_k, where 
! R_z,pi is rotation by pi about the z-axis.  For cases where V_KS lacks R_z,pi symmetry, 
! we'll need to solve explicitly for u_{-k} and then use it in this routine.  

  use constants
  use parallel_data_module
  use grid_module
#ifdef MPI
      use mpi 
#endif

  implicit none

! The grid points and mappings
  type(grid_data), intent(in) :: grid
! A version of grid%indexg() that actually is an inverse of grid%kx, ky, and kz  
  integer, intent(in) :: kInverse(grid%nxmin:grid%nxmax, grid%nymin:grid%nymax, grid%nzmin:grid%nzmax)    
! The parallel environment
  type(parallel_data), intent(in) :: parallel
! Eigenvectors of H + i Gamma
  complex(dpc), intent(in) :: evecall(parallel%ndim)
  complex(dpc), intent(in), optional :: evecall2(parallel%ndim)
! Computed normalization constant 
  complex(dpc), intent(out) :: tau
#ifdef MPI
  integer, intent(inout) :: mpinfo   
#endif

! Work variables
  integer :: igrid, jgrid, kx, ky   
  complex(dpc) :: tau2

  tau = zzero 
  tau2 = zzero

#ifdef DEBUG  
write(9,*) 'xrange', grid%nxmin, grid%nxmax
write(9,*) 'yrange', grid%nymin, grid%nymax
write(9,*) 'zrange', grid%nzmin, grid%nzmax
write(9,*) 'shifts', grid%shift(1), grid%shift(2), grid%shift(3)
call myflush(9)  
#endif 
  if (parallel%iamgmaster) then
  ! Evecall has already been gathered to parallel%group_master, so only perform calculation on parallel%group_master
	do igrid=1,parallel%ndim
		kx = -(grid%kx(igrid) + nint(two*grid%shift(1)) )
		ky = -(grid%ky(igrid) + nint(two*grid%shift(2)) )

call myflush(9)
! Ensure kx, ky are within the DFT supercell 
! We don't check for multiple lattice lengths away since the original point k must be within the supercell
		if (kx < grid%nxmin + grid%norder) then
#ifdef DEBUG  
			write(9,*) '  kx', kx, '-> ', kx + grid%n1
#endif
			kx = kx + grid%n1
		elseif (kx > grid%nxmax - grid%norder) then 
#ifdef DEBUG  
			write(9,*) '  kx', kx, '-> ', kx - grid%n1
#endif
			kx = kx - grid%n1
		endif 
		if (ky < grid%nymin + grid%norder) then
#ifdef DEBUG  
			write(9,*) '  ky', ky, '-> ', ky + grid%n2
#endif
			ky = ky + grid%n2
		elseif (ky > grid%nymax - grid%norder) then 
#ifdef DEBUG  
			write(9,*) '  ky', ky, '-> ', ky - grid%n2
#endif
			ky = ky - grid%n2
		endif 
		jgrid = kInverse(kx, ky, grid%kz(igrid))
call myflush(9)
		if (present(evecall2)) then 
			tau = tau + evecall(igrid)*evecall2(jgrid)  
			tau2 = tau2 + evecall(igrid)*evecall2(igrid)
		else 
			tau = tau + evecall(igrid)*evecall(jgrid)  
			tau2 = tau2 + evecall(igrid)*evecall(igrid)
		endif 
	enddo 
  endif   ! Only perform calculation on parallel%group_master

#ifdef MPI
  call MPI_BCAST(tau, 1, MPI_DOUBLE_COMPLEX, parallel%group_master, parallel%group_comm, mpinfo)
#endif  
  
#ifdef DEBUG  
write(9,*) 'normalize_kpt(): tau = ', tau, ', tau2 =', tau2
call myflush(9)
#endif

end subroutine normalize_kpt



subroutine align_epair_kpt( nstate, eval, evalNoRz, parallel, bestMatchInd ) 
!subroutine align_epair_kpt( nstate, eval, evec, evalNoRz, evecNoRz, parallel, bestMatchInd ) 
! Use (least squares diff of evalues) / nstate to sort -k epairs in agreement with +k  
	 
  use constants
  use parallel_data_module
  use conductance_module
#ifdef MPI
  use mpi
#endif

  implicit none
  
! Number of solved-for eigenpairs, and size of evec, eval arrays
  integer, intent(in) :: nstate 
! The parallel environment
  type(parallel_data), intent(in) :: parallel
! The eigenvectors of H + i Gamma
!  complex(dpc), intent(in) :: evec(parallel%ldn, nstate), evecNoRz(parallel%ldn, nstate)
! Eigenvalues of H + i Gamma
  complex(dpc), intent(in) :: eval(nstate), evalNoRz(nstate)
  integer, intent(out) :: bestMatchInd(nstate)

  real(dp), parameter :: collisionCutoff = 7.3d0

 
! A matrix giving the 'distance metric' between each eigenpair in +k and each eigenpair in -k  
  !real(dp) :: distance(nstate, nstate)
  real(dp) :: distance(nstate), bestMatch(nstate), nextBestMatch(nstate)
  integer :: nextBestMatchInd(nstate), numRefs(nstate) 
  integer :: alcstat, i, k, HoleSpurious, iKeep  
  real(dp) :: bestMatchKeep
  
  numRefs(:) = 0
  do i = 1, nstate
	distance(:) = abs( eval(i) - evalNoRz(:) )
	
	bestMatchInd(i) = minloc(distance, 1)
	bestMatch(i) = distance(bestMatchInd(i)) 
	numRefs(bestMatchInd(i)) = numRefs(bestMatchInd(i)) + 1
  enddo 	
		
  
!!!!! Find collisions and resolve them 
! Possibilities: (i) Degeneracy (i.e., best match not unique & agreement of +-k sides is actually normal)  (ii) Hole or spurious epair on +k side  (iii) Hole or spurious epair on -k side 
! Note a hole on a given side is like spurious epair because it is filled by the next higher epair from the top of the list, which is missing from the opposite side's list.  In fact, a spurious epair also means a hole because the *top* epair will be pushed off the top end of the list.  
! Note possibilities (ii) and (iii) should really be treated the same, as we're not going back to solve for a hole on +k side, so we'll end up treating this epair on -k side *as if* it's spurious.  So really we only need to treat case (i) separately from (ii)-(iii) 
! It will likely be more of a challenge to differentiate reliably between (i) and (ii)-(iii) than to deal with each case !  Perhaps the evector biorthogonal (dot) product can help with this?   
! But maybe differentiating these cases is not truly so bad because in case (i), there are say two epairs on -k side that both have the same best match on +k side, but then there is another epair on +k side that could potentially match (either) one of them.  This epair goes without any good match during the collision, & hence could be used to resolve the collision.  Whereas in cases (ii)-(iii), attempting to match both epairs on -k side to the +k side will worsen all the matches of everything.  So perhaps the best way to judge is look at the global goodness-of-match picture.  

! Heuristic: find next best match for each epair on -k side involved in the collision.  If any epair on +k side is a not much worse match (say no more than 10x the distance 'metric'), and is also a better match than any *other* epair (not involved in collision), then accept this as case (i).  If not, treat as (ii)-(iii)  ?   

  do i = 1, nstate
	if ( numRefs(i) .neq. 1 ) then
	
#ifdef DEBUG  
		write(9,*) 'align_epair_kpt: Collision ', i, numRefs(i)
#endif
		nextBestMatchInd(:) = 0
		nextBestMatch(:) = zero 
		HoleSpurious = 0
	
		! Find next-best match for each -k epair pointing to the collided +k epair  
		iKeep = -1 
		bestMatchKeep = maxval(distance) + 500 
		
		do k = 1, nstate
			if (bestMatchInd(k) == i) then
				distance(:) = abs( eval(k) - evalNoRz(:) )
				distance(i) = distance(1) + 500 
			
				nextBestMatchInd(k) = minloc(distance, 1)
				nextBestMatch(k) = distance(nextBestMatchInd(k)) 
#ifdef DEBUG  
				write(9,*) 'Next-best match: ', k, nextBestMatchInd(k), nextBestMatch(k), nextBestMatch(k) / bestMatch(k)
#endif
				
				distance(:) = abs( eval(:) - evalNoRz(nextBestMatchInd(k)) )
				if (nextBestMatch(k) / bestMatch(k) > collisionCutoff  .or. &
					nextBestMatch(k) > minval(distance)) then  ! Note distance(:) has been redefined since it was used to set nextBestMatch(k)
						HoleSpurious = HoleSpurious + 1
						if (iKeep < 1 .or. bestMatch(k) < bestMatchKeep) then 
							bestMatchKeep = bestMatch(k)
							if (iKeep > 0) bestMatchInd(iKeep) = -1
							iKeep = k
						else 
							bestMatchInd(k) = -1
						endif
				endif
					
			endif 
		enddo 
		
		! Need to tweak this decision to be a little less binary.  Each hole/spurious pair is 'independent,' i.e. each one can be thrown out on the merits of its own decision.  For whichever one or more epairs in -k we don't throw out, we should decide *collectively* which epairs in -k go with which in +k.  We want to optimize the fit of the list *as a whole*.  
		if (HoleSpurious > 0) then 
#ifdef DEBUG  
			write(9,*) 'Identify as hole / spurious epair' 			
			write(9,*) 'Keeping ', iKeep, bestMatchKeep 
#endif
			
			
		else 
#ifdef DEBUG  
			write(9,*) 'Identify as degeneracy' 
#endif
			k = minloc(nextBestMatch, 1)
#ifdef DEBUG  
			write(9,*) 'Best: ', k, nextBestMatchInd(k), nextBestMatch(k) 
#endif
		endif

	endif 
  enddo 	
  
end subroutine align_epair_kpt



#ifdef MPI
subroutine basis_change_CAPs(nstate, parallel, cond, evec, KPtNoRz, evecNoRz, bestMatchInd, GamLR, mpinfo)
#else
subroutine basis_change_CAPs(nstate, parallel, cond, evec, KPtNoRz, evecNoRz, bestMatchInd, GamLR)
#endif
! Compute product GamLR OF CAPs in diagonal basis of (H + i Gamma)

  use constants
  use parallel_data_module
  use conductance_module
#ifdef MPI
      use mpi
#endif

  implicit none

! Number of solved-for eigenpairs, and size of evec, GamLR arrays
  integer, intent(in) :: nstate 
! The parallel environment
  type(parallel_data), intent(in) :: parallel
! Parameters defining the conductance calculation to perform 
  type(conductance_param), intent(inout) :: cond
! The eigenvectors of H + i Gamma
  complex(dpc), intent(in) :: evec(parallel%ldn, nstate), evecNoRz(parallel%ldn, nstate)
! Flag for having explicit solutions for -k points (because R_z,pi symmetry is absent) 
! passed in evecNoRz (otherwise this should be an unallocated array) 
  logical, intent(in) :: KPtNoRz
  integer, intent(in) :: bestMatchInd(nstate)
#ifdef MPI
  integer, intent(inout) :: mpinfo 
#endif
! To hold the CAPs in the diagonal basis of H + i Gamma
  complex(dpc), dimension (nstate,nstate), intent(out) ::  GamLR  


  integer :: alcstat, igrid, jgrid, kgrid
  complex(dpc) :: tau1, tau2
  !  For transforming CAPs to eigenbasis of H + i Gamma
  complex(dpc), dimension(nstate) :: r, w, s, u, rAll, wAll, sAll, uAll  


  do igrid = 1, nstate

     r(:) = zzero
     s(:) = zzero 
     u(:) = zzero 
     w(:) = zzero
	 
	 
	 ! We only explicitly compute half the transformed CAPs (jgrid <= igrid, 
	 ! representing a lower triangle for GamLR).  We
	 ! make use of the CAPs' Hermiticity to derive the full transformed matrices.   
     do jgrid = 1, igrid  
     
#ifdef DEBUG  
     if (kPtNoRz)  write(9,*) 'basis_change():', igrid, jgrid, bestMatchInd(igrid), bestMatchInd(jgrid)   ! DEBUG  
#endif

     	do kgrid = 1, parallel%mydim*parallel%mxwd
	   tau1 = conjg(evec(kgrid, igrid)) * evec(kgrid, jgrid) 

	   !!!!! If no R_z,pi symm, add explicit matrix elements for -k points
	   if (KPtNoRz) then  
	   		if (bestMatchInd(igrid) < 0  .or.  bestMatchInd(jgrid) < 0) then
	   			tau1 = zzero
	   		else
				tau2 = conjg(evecNoRz(kgrid, igrid)) * evecNoRz(kgrid, jgrid)
			endif 
	   endif 

	   r(jgrid) = r(jgrid) + tau1 * cmplx( cond%gammal(kgrid), 0.d0)
	   if (KPtNoRz) s(jgrid) = s(jgrid) + tau2 * cmplx( cond%gammal(kgrid), 0.d0)
	   u(jgrid) = u(jgrid) + tau1 * cmplx( cond%gammar(kgrid), 0.d0)
	   if (KPtNoRz) w(jgrid) = w(jgrid) + tau2 * cmplx( cond%gammar(kgrid), 0.d0)

	   
	enddo
     enddo

     ! Reduce GamLR (finish the summation over kgrid)
#ifdef MPI
     call MPI_ALLREDUCE(r(1:igrid), rAll(1:igrid), igrid, MPI_DOUBLE_COMPLEX, MPI_SUM, parallel%group_comm, mpinfo)
     call MPI_ALLREDUCE(u(1:igrid), uAll(1:igrid), igrid, MPI_DOUBLE_COMPLEX, MPI_SUM, parallel%group_comm, mpinfo)
	 GamLR(igrid, 1:igrid) = rAll(1:igrid)*uAll(1:igrid)
	 if (kPtNoRz) then
	      call MPI_ALLREDUCE(s(1:igrid), sAll(1:igrid), igrid, MPI_DOUBLE_COMPLEX, MPI_SUM, parallel%group_comm, mpinfo)
		call MPI_ALLREDUCE(w(1:igrid), wAll(1:igrid), igrid, MPI_DOUBLE_COMPLEX, MPI_SUM, parallel%group_comm, mpinfo)
		GamLR(igrid, 1:igrid) = (((rAll(1:igrid)+sAll(1:igrid)) * (uAll(1:igrid)+wAll(1:igrid))) &
			- ((rAll(1:igrid)-sAll(1:igrid)) * (uAll(1:igrid)-wAll(1:igrid)))) / cmplx(4.d0, zero)
	 endif
#else
	if (kPtNoRz) then
		GamLR(igrid, 1:igrid) = (((r(1:igrid)+s(1:igrid)) * (u(1:igrid)+w(1:igrid))) &
			- ((r(1:igrid)-s(1:igrid)) * (u(1:igrid)-w(1:igrid)))) / cmplx(4.d0, zero)
	else
		GamLR(igrid, 1:igrid) = r(1:igrid)*u(1:igrid)
	endif 
#endif
	! Explicitly set the full transformed CAPs using their Hermiticity
	GamLR(1:(igrid-1), igrid) = conjg(GamLR(igrid, 1:(igrid-1)))
	GamLR(igrid, igrid) = real(GamLR(igrid, igrid))
  enddo

end subroutine basis_change_CAPs



subroutine compute_TE(nstate, GamLR, cond, eval, spinID, fileUnit, ipr)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Compute Transmission T(E) at energies E based on the product GamLR of CAPs  
! expressed in the diagonal basis of (H + i Gamma) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!     

	use constants
	use conductance_module

	! Number of solved-for eigenpairs, and size of the input arrays
	integer, intent(in) :: nstate
    ! The CAPs in the diagonal basis of H + i Gamma
    complex(dpc), dimension (nstate,nstate), intent(in) ::  GamLR  
	! Parameters defining the conductance calculation to perform 
    type(conductance_param), intent(inout) :: cond
    ! Eigenvalues of H + i Gamma
    complex(dpc), intent(in) :: eval(nstate)
	! File unit to write to for TE.dat  
	integer, intent(in) :: fileUnit
	! Spin identifier 
	character(len=25), intent(in) :: spinID
	! Print-out level 
    integer, intent(in) :: ipr

	integer :: i, j, k
      complex(dpc) :: TE 

	open(fileUnit,file='TE.'//trim(spinID)//'.dat',form='formatted')
	write(fileUnit,*) '    E (Ry)                      E - E_F (eV)', &
		'                      Re{T(E)}                      Im{T(E)}' 
     	write(fileUnit,*) '----------------------------------------------------------------------'

! T(E) calculation: loop over energies E
      EnergyLoop: do k=1, cond%NE

      	if (ipr >= 2) then  
			write(9,*) "----------------------------------------------------------------------"
			write(9,*) "Calculation of T(E) at energy point ", k, " out of ",cond%NE," points."
			write(9,*) "----------------------------------------------------------------------"
			call myflush(9)
		endif 
	  
		! Set T(E) = 0, to be summed 
      	TE = zzero


      	do i=1, nstate
            	do j=1, nstate
            		TE = TE + GamLR(i, j) / &
	       	  ((cmplx(cond%Espectrum(k),0.d0) - eval(j)) * (cmplx(cond%Espectrum(k),0.d0) - conjg(eval(i))))

            	enddo
            enddo

 	   	TE = TE * cmplx(4.d0, 0.d0) 

		! Write Transmission TE to disk
            write(fileUnit,*) cond%Espectrum(k), (cond%Espectrum(k) - cond%efermi) * rydberg, &
            	real(TE), aimag(TE)
            call myflush(fileUnit)
!	   close(fileUnit)

      enddo EnergyLoop
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! End loop over energy
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	close(fileUnit)
	
end subroutine compute_TE




subroutine zarpk_gamma_diag(kplp, Hks_cplx, RetAdv, parallel,nstate, nec, &

     cond, maxmvs,eval,evec,tol,resid_restart,ipr,info)

  use constants
  use parallel_data_module
  use conductance_module
  implicit none
  include 'arpkdbg.h'
  !
  ! Input/Output variables:
  !
  ! parallel computation related data
  type (parallel_data), intent(in) :: parallel

  integer, intent(in) :: ipr, nstate, RetAdv ! Retarded vs. Advanced Green's function
  integer, intent(inout) :: maxmvs

  integer, intent(in) :: kplp
  logical, intent(in) :: Hks_cplx ! Value of elec_st%cplx -- is Hks itself complex (because of k-points)?

  integer, intent(inout) :: info

  real(dp), intent(in) :: tol
  complex(dpc), intent(out) :: eval(nstate)
  complex(dpc), intent(inout) :: evec(parallel%ldn*parallel%mxwd,nstate)

  complex(dpc), intent(inout) :: resid_restart(parallel%ldn*parallel%mxwd)
  !
  ! Work variables:
  !
  integer :: ndim, ldn, i

  type(conductance_param), intent(in) :: cond

  ! arpack scalars
  character (len=1) :: bmat
  character (len=2) :: which
  integer :: ido, nev, ncv, lworkl, j,jj
  integer :: k
  integer, intent(inout) :: nec

  logical :: rvec
  complex(dpc) :: sigma, xscal

  ! arpack variables

  ! change (adi):
  complex(dpc), allocatable :: vbasis(:,:)
  real(dp), dimension(:), allocatable :: rwork
  complex(dpc), dimension(:), allocatable :: eval_loc
  complex(dpc), dimension(:), allocatable :: resid,workd
  complex(dpc), dimension(:), allocatable :: workl

  complex(dpc), dimension(:), allocatable :: workev


  logical, allocatable :: select1(:)
  integer, allocatable :: indx(:)
  integer iparam(11)

  integer ipntr(14)



  ! allocation check
  integer alcstat

  !-------------------------------------------------------------------




  !
  ! Initialize the arpack variables.
  !
  nev = nstate
  ncv = 2*nstate+5

  lworkl = ncv*(3*ncv+5)



  ndim = parallel%mydim*parallel%mxwd
  ldn  = parallel%ldn*parallel%mxwd
  eval(:) = zzero



  allocate(vbasis(ndim, ncv),stat=alcstat)
  call alccheck('arpk_vbasis',ndim*(ncv),alcstat)

  allocate(workl(lworkl),stat=alcstat)
  call alccheck('arpk_workl',lworkl,alcstat)
  allocate(workd(3*ndim),stat=alcstat)
  call alccheck('arpk_workd',3*ndim,alcstat)
  allocate(resid(ndim),stat=alcstat)
  call alccheck('arpk_resid',ndim,alcstat)

  allocate(eval_loc(ncv))
  allocate(rwork(ncv))

  allocate(workev(2*ncv))
  workev(:) = zzero

  allocate(select1(ncv))

  vbasis(:,:) = zzero
  rwork(:) = zero
  workl(:) = zzero
  workd(:) = zzero
  eval_loc(:) = zzero
  resid(:) = zzero
  iparam(:) = 0
  ipntr(:) = 0

  bmat = 'I'         !standard eigenvalue problem

  which = 'SM'       !compute the smallest-magnitude modes 




  call zcopy(ndim,resid_restart,1,resid,1)

  ido = 0            !integer used to control reverse communication

  iparam(1) = 1      !choose the "exact-shift" strategy for restart
  iparam(3) = nev*90 !maxium number of implicit restart allowed
  iparam(4) = 1      !block size, current ARPACK is single vector version
  iparam(7) = 1      !regular mode without invert or shift-invert

  ! Initialize some variables included from debug.h, only useful when ipr > 1.
  if (ipr >= 2) then
     ndigit = -3
     logfil = 6
     msgets = 0
     msaitr = 0
     msapps = 0
     msaupd = 0
     msaup2 = 0
     mseigt = 0
     mseupd = 1         
  endif

  ! Main loop (Reverse Communication)

  ArpackLoop: do


#ifdef MPI
     call My_pznaupd (parallel%group_comm, &
#else
     call My_znaupd ( &
#endif
          ido, bmat, ndim, which, nev, tol, resid, &
          ncv, vbasis, ndim, iparam, ipntr, &
          workd, workl, lworkl, rwork, info)
     if (ido /= -1 .and. ido /= 1) exit
!     call arpk_zmatvec(1,1,workd(ipntr(1)), &
!         ndim,workd(ipntr(2)),ndim)
    call arpk_condmatvec(kplp, Hks_cplx, RetAdv, workd(ipntr(1)), ndim, workd(ipntr(2)), &
    	 	ndim, cond%gammafull, real(cond%shift), parallel) 

    ! For use with k-points: if RetAdv = -1 (denoting case of explicit solution for -k),  
    ! take complex conjugate of the output from condmatvec()      
    ! if (RetAdv < 1) workd(ipntr(2):(ipntr(2)+ndim-1)) = conjg(workd(ipntr(2):(ipntr(2)+ndim-1)))


    if (modulo(iparam(9),100) .eq. 0 ) then
#ifdef MPI
	write(9,*) 'Arpack-PZNaupD iter ', iparam(9), ': '
#else
	write(9,*) 'Arpack-ZNaupD iter ', iparam(9), ': '
#endif
	write(9,*) '  Nconv = ', iparam(5), ' ido = ', ido, ', info = ', info
	call myflush(9)
    endif

     if (iparam(9) >= maxmvs) then
     	write(9,'(/,a,/)') ' Maximum number of iterations reached.'
     	exit
     endif
  enddo ArpackLoop

  write(9,*) '# iterations: ', iparam(9)
  write(9,*) 'mxwd: ', parallel%mxwd


  if ( info == 1) then
     write(9,'(/,a,/)') ' Maximum number of iterations reached.'
  else if ( info == 3) then
     write(9,'(/,a,a,/)') ' No shifts could be applied during', &
          ' implicit Arnoldi update, try increasing NCV.'
  end if


  write(9,*) 'Before pzneupd'


  if ( info < 0 ) then
     write(9,'(/,a,i5)') ' Error with _saupd, info = ', info
     write(9,'(a,/)') ' Check documentation in _saupd '
  else

     ! Postprocess with dseupd.

     rvec = .true.
#ifdef MPI
     call My_pzneupd(parallel%group_comm, &
#else
     call zneupd( &
#endif
         rvec, 'A', select1, eval_loc, vbasis, ndim, &
         sigma,workev,bmat, ndim, which, nev, tol,  &
         resid, ncv,vbasis, ndim, iparam, ipntr, &
         workd, workl, lworkl, rwork, info)
     resid_restart(:) = zzero

     do jj= 1,nev
        do j = 1, ndim
           resid_restart(j) = resid_restart(j)+vbasis(j,jj)
        enddo
     enddo
     xscal = zone/real(nev,dp)
     call zscal(ndim,xscal,resid_restart,1)

     if ( info /= 0) then
        write(9,*)
        write(9,*) ' Error with _seupd, info = ', info
        write(9,*) ' Check the documentation of _seupd. '
        write(9,*)
     end if

     ! Print additional convergence information.

     if (ipr >= 2) then

        write(9,*) ' '
        write(9,*) ' _SDRV1 '
        write(9,*) ' ====== '
        write(9,*) ' '
        write(9,*) ' Size of the matrix is ', ndim
        write(9,*) ' The number of Ritz values requested is ', nev
        write(9,*) ' The number of Arnoldi vectors generated', &
             ' (NCV) is ', ncv
        write(9,*) ' What portion of the spectrum: ', which
        write(9,*) ' The number of converged Ritz values is ',  &
             iparam(5)
        write(9,*) ' The number of Implicit Arnoldi update', &
             ' iterations taken is ', iparam(3)
        write(9,*) ' The number of OP*x is ', iparam(9)
        write(9,*) ' The convergence criterion is ', tol
        write(9,*) ' '

     end if

  end if

  write(9,*) 'After pzneupd'
 
  ! evec <-  vbasis
  nec = iparam(5)
  maxmvs = iparam(9)
       
  ! Make sure eigenvalues are ordered in ascending order of value.
  write(9,*) 'Before allocate'
  allocate(indx(nec))
  write(9,*) 'After allocate'
  write(9,*) 'Eigenvalues: '
  do j = 1, nec
     indx(j) = j
     rwork(j) = abs(eval_loc(j)) ! real(eval_loc(j),dp)
  enddo

  write(9,*) 'Before sortreal'
  call sortreal(which, nec, rwork, indx, zero)

  write(9,*) 'After sortreal'

  evec(:,:) = zzero
  do j=1, min(ncv,nstate)
     eval(j) = eval_loc(indx(j)) + real(cond%shift) ! rwork(j)
	 if (RetAdv < 1) eval(j) = conjg(eval(j))
     write(9,*) j, eval(j)
     do i = 1, ndim
        evec(i,j) = vbasis(i,indx(j))
		if (RetAdv < 1) evec(i,j) = conjg(evec(i,j))
     enddo
  enddo
  call myflush(9)
  deallocate(indx)
  !
  if (allocated(vbasis)) deallocate(vbasis)
  if (allocated(workl)) deallocate(workl)
  if (allocated(workd)) deallocate(workd)
  if (allocated(resid)) deallocate(resid)
  if (allocated(eval_loc)) deallocate(eval_loc)
  if (allocated(rwork)) deallocate(rwork)
  if (allocated(workev)) deallocate(workev)
  if (allocated(select1)) deallocate(select1)


end subroutine zarpk_gamma_diag
!===================================================================





subroutine  arpk_condmatvec(kplp, Hks_cplx, RetAdv, v, ldnv, w, ldnw, gammafull, EShift, parallel) 


!  use parsec_global_data
  use parallel_data_module
  use matvec_module
  implicit none
  !
  !  Input/Output variables:
  !
  integer, parameter :: blksize = 1

  type(parallel_data), intent(in) :: parallel
  integer, intent(in) ::  kplp, ldnv, ldnw ! , blksize
  integer, intent(in) ::  RetAdv	! Retarded vs. Advanced Green's fn
  logical, intent(in) ::  Hks_cplx  ! Value of elec_st%cplx -- is Hks itself complex (because of k-points)?
  complex(dpc), intent(in) ::  v(ldnv, blksize)
  complex(dpc), intent(out) ::  w(ldnw, blksize)
  real(dp), intent(in)  :: gammafull(parallel%ldn)
  real(dp), intent(in) :: EShift
  !  Work variables:
  !
  complex(dpc) ::  v_loc(parallel%ldn*parallel%mxwd) !, blksize)
  complex(dpc) ::  w_loc(parallel%ldn*parallel%mxwd) !, blksize)
  integer :: ii, jj, ll, mm, alcstat
  !real(dp),  pointer :: myvec(:)

  !-------------------------------------------------------------------


  !allocate(myvec(parallel%nwedge+1),stat=alcstat)
  v_loc(:) = zzero
  do ii = 1, blksize
     do jj = 1, parallel%mxwd
        ll = (jj-1)*parallel%ldn
        mm = (jj-1)*parallel%mydim
        v_loc(ll+1:ll+parallel%mydim) = -v(mm+1:mm+parallel%mydim,ii)
     enddo
  enddo
  if (Hks_cplx) then
    call zcondmatvec(kplp, gammafull, EShift, v_loc, w_loc, RetAdv, parallel) 
  else
    call condmatvec(kplp, gammafull, EShift, v_loc, w_loc, RetAdv, parallel) 
  endif
!  call condmatvec(kplp,blksize,v_loc,w_loc)

  w(:,:) = zzero
  do ii = 1, blksize
     do jj = 1, parallel%mxwd
        ll = (jj-1)*parallel%ldn
        mm = (jj-1)*parallel%mydim
        w(mm+1:mm+parallel%mydim,ii) = w_loc(ll+1:ll+parallel%mydim)
     enddo
  enddo
!   call simp_matvec(kplp,blksize, v, w, myvec)

!   if(associated(myvec)) deallocate(myvec)


end subroutine arpk_condmatvec



subroutine Change_MPI_Gps_Cond(cond, parallel, elec_st, grid, pbc, pot, solver, rsymm, clust, p_pot, u_pot, nloc_p_pot)

      use parallel_data_module
      use conductance_module
      use electronic_struct_module
      use grid_module
      use pbc_module
      use symmetry_module
      use eigen_solver_module
      use potential_module
      use non_local_psp_module
      use pseudo_potential_module
      use cluster_module
      use constants 
	  !use parsec_global_data

#ifdef MPI
      use mpi
#endif

!	Parameters defining the conductance calculation to perform 
      type(conductance_param), intent(inout) :: cond
!     The parallel environment
      type(parallel_data), intent(inout) :: parallel
!     Electronic structure information
      type(electronic_struct), intent(inout) :: elec_st
	  !  grid related data
	  type (grid_data), intent(inout) :: grid
	  !  periodic boundary conditions data
	  type (pbc_data), intent(inout) :: pbc
	  ! solver related data
	  type(eigen_solver), intent (inout) :: solver
	  !	Potential information
      type(potential), intent(inout) :: pot
	  !  symmetry operations in reduced (Abelian) group
	  type (symmetry), intent(in) :: rsymm
	  !  the cluster
	  type (cluster), intent(in) :: clust
	  !  grid related data
	  type (pseudo_potential), intent(inout) :: p_pot
	  !  non local pseudo_potential related data
	  type (nonloc_pseudo_potential), intent(inout) :: nloc_p_pot
	  !  non local pseudo_potential related data
	  type (nonloc_pseudo_potential), intent(inout) :: u_pot

!     Work variables
      integer :: ii, igrid, isp, kplp, irp, mpinfo, NumEigensolutionsPerSpin  
      real(dp), dimension (parallel%ndim) :: adiag
      real(dp), dimension (parallel%ndim, elec_st%nspin) :: vnew  	  
	  integer, dimension (parallel%ndim) :: kx, ky, kz 
	  ! A version of grid%indexg() that actually is an inverse of grid%kx, ky, and kz  
	  integer :: kInverse(grid%nxmin:grid%nxmax, grid%nymin:grid%nymax, grid%nzmin:grid%nzmax)   
	
!write(9,*) ' Change_MPI_Gps_Cond(): Before Bcast()'
!call myflush(9)
#ifdef MPI
	call MPI_BCAST(cond%NumGroups,1,MPI_INTEGER,parallel%masterid,parallel%comm,mpinfo)
	call MPI_BCAST(cond%Num_Ignore_KPoints,1,MPI_INTEGER,parallel%masterid,parallel%comm,mpinfo)
	if (cond%Num_Ignore_KPoints > 0) then 
	  if (.not. parallel%iammaster) allocate(cond%Ignore_KPoints(cond%Num_Ignore_KPoints))  
	  call MPI_BCAST(cond%Ignore_KPoints,cond%Num_Ignore_KPoints,MPI_INTEGER, &
		parallel%masterid, parallel%comm, mpinfo)
	endif 
	call MPI_BCAST(cond%Parallelize_Over_SignK,1,MPI_INTEGER,parallel%masterid,parallel%comm,mpinfo)
#endif
!write(9,*) ' Change_MPI_Gps_Cond(): After Bcast()' 
!call myflush(9) 

if(parallel%iammaster) then 
   write(7,*) 'Getting ready to change MPI_groups if required: cond%NumGroups= ', cond%NumGroups    
   call myflush(7)
 endif
   write(9,*) 'Getting ready to change MPI_groups if required: cond%NumGroups= ', cond%NumGroups
   call myflush(9)

 
 if (cond%NumGroups == 0 .or. (cond%NumGroups > 0 .and. cond%NumGroups /= parallel%groups_num) .or. &
		(cond%NumGroups > -1 .and. cond%Parallelize_Over_SignK > -1)) then
	! For this condition to work right, should really have done two things previously: 
	!	(i) If cond_MPI_Groups = MPI_Groups specified in parsec.in, set cond%NumGroups = -1.  
	!	(ii) If cond_MPI_Groups = parallel%groups_num resulting *after* the call to create_group_layout() (defined in structures.f90p), set cond%NumGroups = -1.
	!		This (ii) is a little redundant with the "cond%NumGroups /= parallel%groups_num" branch of the condn.  I should decide whether that redundancy is desirable (making code safer) or not.  
	! Actually, I'm also concerned there is one case where cond%NumGroups = parallel%groups_num = 0, yet the result of automatic setting differs for DFT vs. transport.  That's the case where the number of k-points differs, either because R_z,pi symmetry is missing and the gamma point has a different weight, or because k-points are ignored in the transport calculation but are present in the DFT.  I don't want to make the if condn very complex, but I do need to deal with this case in some way.  
	! 	But maybe I don't really need to worry about this one case separately.  My goal was to have a way to do a trial assignment or "dry run" to check whether applying this target would lead to an actually diff # of MPI gps, and if it wouldn't, don't bother changing # of gps.  So whether or not this "dry run" took place (as a result of this if condition), it still shouldn't affect the actual # of gps.  It would be OK to do the "dry run" every time cond%NumGroups = 0.  
	
	
	! Perform "dry run" to determine whether the new actual # of MPI gps will differ 
	! from current #
    if (cond%NumGroups == 0) then
       !groups_num = rsymm%ntrans*nspin*(kpnum - cond%Num_Ignore_KPoints)
       !cond%NumGroups = rsymm%ntrans*nspin*(kpnum - cond%Num_Ignore_KPoints)
	   !Ignore_symmetry should always be true for conductance calculations, so in principle rsymm%ntrans is always 1  
	   cond%NumGroups = elec_st%nspin*(elec_st%nkpt - cond%Num_Ignore_KPoints)  
    !else
    !   groups_num = cond%NumGroups
    endif
    do ii = cond%NumGroups, 1, -1
       if (mod(parallel%procs_num,ii) == 0 .and. &
            mod(cond%NumGroups,ii) == 0) then  ! This criterion will also need to be updated for the case where # k pts > # PEs 
          cond%NumGroups = ii
          exit
       endif
    enddo

	if(parallel%iammaster) then 
		write(7,*) 'cond%NumGroups= ', cond%NumGroups, 'parallel%groups_num= ', parallel%groups_num
		call myflush(7)
	endif 
#ifdef MPI 
	if ((.not. elec_st%cplx) .or. (elec_st%nkpt < 1) .or. (cond%RzSymm)) cond%Parallelize_Over_SignK = -1
	if ((cond%NumGroups /= parallel%groups_num) .or. (cond%Parallelize_Over_SignK > -1)) then
		parallel%groups_num = cond%NumGroups
		! Call MPI destructors MPI_group_free() and MPI_comm_free()  
		if (parallel%iamgmaster) then 
			write(7,*) 'parallel%groups_num set to ', parallel%groups_num 
			call myflush(7)
		endif 
!		write(9,*) ' Before gatherv()'
!		call myflush(9)
		call MPI_ALLGATHERV(solver%adiag, parallel%mydim, MPI_DOUBLE_PRECISION, adiag, &
			parallel%irows(1:parallel%procs_num)-parallel%irows(0:parallel%procs_num-1), &
			parallel%irows(0:parallel%procs_num-1)-1,MPI_DOUBLE_PRECISION, parallel%group_comm, mpinfo)
		do isp = 1, pot%nspin
			call MPI_ALLGATHERV(pot%vnew(:,isp), parallel%mydim, MPI_DOUBLE_PRECISION, vnew(:,isp), &
				parallel%irows(1:parallel%procs_num)-parallel%irows(0:parallel%procs_num-1), &
				parallel%irows(0:parallel%procs_num-1)-1,MPI_DOUBLE_PRECISION, parallel%group_comm, mpinfo)
		enddo 
!		if(parallel%iammaster) then  
!			write(7,*) 'call create_group_layout(parallel, cond%NumGroups)'  
!			call myflush(7)  
!		endif 
		kx(1:parallel%ndim) = grid%kx(1:parallel%ndim)
		ky(1:parallel%ndim) = grid%ky(1:parallel%ndim)
		kz(1:parallel%ndim) = grid%kz(1:parallel%ndim) 
! 		
		! Set parallel%mydim based on irows, as in l. 343 of setup() 
		! Reset parallel%neibs to its global form (as in l. 157-161 and 187-294 of setup()), since this is what comm_neigh() expects on input.  
		! call grid_partition(grid,pbc,rsymm,parallel,ierr)   ! Or new routine for "PART2: Partitioning The Domain," l. 368-501 of grid_partition.f90   
		! Be sure to set parallel%ldn to an appropriate value (l. 338 of grid_partition.f90).  Some (or many?) arrays may need to be reallocated & maybe gathered/broadcast -- esp converged solver%adiag (V_KS), rho, KS orbitals 
		!      Note that solver%adiag is set initially in eigval() to pot%vnew + diagonal part of KE.  Its size is allocated as pot%ndim.  This nomenclature is slightly inconsistent / confusing because pot%ndim = parallel%mydim (= local # of grid pts on each core), whereas parallel%ndim = total # of grid pts.   
		!			(i) MPI_Gather() the actual data structures (e.g., adiag) to some temporary array on the master PE as target.  (ii) Free & reallocate the actual data structures on all PEs.  (iii) MPI_Scatter() the temporary arrays into the reallocated actual data structures as targets.  
		!				To refine this process: (i) MPI_Gather() adiag etc. to temporary arrays.  (ii) Free actual data structs like adiag etc, as well as irows, kx etc.  (iii) Call grid_set_wedge() and grid_set_ist() to reallocate actual data structs kx etc.  Also reallocate adiag, etc.  (iv) MPI_Scatter() the temporary arrays into the reallocated actual data structs as targets. 
		!					**Note this is a very minimal, basic algorithm for reallocatg / redistributg the grids & volumetric data.  It's not robust enough to eventually handle the general case.  But it could serve for now as a proof-of-principle test, to use e.g. with the H Chain, and it could eventually serve as a starting point for a more general algorithm.  
!		write(9,*) ' Before deallocate()'
!		call myflush(9)
		! Deallocate the vars that will get allocated in the call to grid_partition() 
		!if (associated (grid%indexg)) deallocate(grid%indexg)
		!if (associated (grid%indexw)) deallocate(grid%indexw)
		!write(9, *) ' After dealloc grid%index'
		!call myflush(9)
		!if(grid%hartree_neibs_flag) deallocate (grid%neibs_index)
		!deallocate(grid%rindex)
		!deallocate(grid%rtrans) 
		!deallocate(grid%fx) 
		!deallocate(grid%fy) 
		!deallocate(grid%fz)
 		!write(9,*) ' After dealloc grid%fxyz'
		!call myflush(9) 
		if (grid%hartree_neibs_flag) then 
			deallocate(grid%neibs_fx)
			deallocate(grid%neibs_fy)
			deallocate(grid%neibs_fz) 
		endif 
!		write(9,*) ' After dealloc grid%neibs'
!		call myflush(9)
		!deallocate(grid%kx)
		!deallocate(grid%ky)
		!deallocate(grid%kz)
		!write(9,*) ' After dealloc grid%kxyz'
		!call myflush(9) 
		deallocate(grid%ist) 
		deallocate(parallel%irecvp)
		deallocate(parallel%jsendp) 
		deallocate(parallel%neibs) 
		deallocate(parallel%sendcounts)
		deallocate(parallel%sources)
		deallocate(parallel%sdispls)
		!deallocate(parallel%source_weights)
		deallocate(parallel%recvcounts)
		deallocate(parallel%destinations)
		deallocate(parallel%rdispls)

	!                if(parallel%iammaster) then
	!                        write(7,*) 'call create_group_layout(parallel, cond%NumGroups)'
	!                        call myflush(7)
	!                endif
	!                call MPI_BARRIER(parallel%comm,mpinfo)
	!                call create_group_layout(parallel, cond%NumGroups)
			
!			write(9,*) ' Before deallocate solver%adiag '
!			call myflush(9) 
			
!			write(9,*) ' Before setup()'
!			write(9,*) '    irows: ', parallel%irows(0:parallel%procs_num)
!			write(9,*) '    mydim: ', parallel%mydim, ' ldn: ', parallel%ldn 
!			call myflush(9)
			call MPI_BARRIER(parallel%comm,mpinfo)
			call setup(grid,pbc,rsymm,parallel,elec_st%nspin,elec_st%nkpt,mpinfo)
			call MPI_BARRIER(parallel%comm,mpinfo)
!			write(9,*) ' After setup()'
!			write(9,*) '    irows: ', parallel%irows(0:parallel%procs_num)
!			write(9,*) '    mydim: ', parallel%mydim, ' ldn: ', parallel%ldn
!			call myflush(9)  
			
			deallocate(solver%adiag)
			deallocate(pot%vnew)
!			write(9,*) ' Before allocate solver%adiag '
!			call myflush(9)
			allocate(solver%adiag(parallel%ldn), stat=mpinfo)
			call alccheck('solver%adiag',parallel%ldn,mpinfo)
			pot%ndim = parallel%mydim  
			allocate(pot%vnew(pot%ndim, pot%nspin), stat=mpinfo)
			call alccheck('pot%vnew', pot%ndim*pot%nspin, mpinfo)
		!allocate(parallel%jsendp(0:parallel%group_size), stat=mpinfo)
		!call alccheck('parallel%jsendp', parallel%group_size+1, mpinfo)
        !        allocate(parallel%irecvp(0:parallel%group_size), stat=mpinfo)
        !        call alccheck('parallel%jsendp', parallel%group_size+1, mpinfo)
		!allocate(grid%ist(0:parallel%group_size), stat=mpinfo)
		!call alccheck('grid%ist',parallel%group_size+1,mpinfo)
		!allocate(parallel%neibs((6+2*grid%lap_dir_num)*grid%norder,parallel%ldn), stat=mpinfo)
		!call alccheck('parallel%neibs',(6+2*grid%lap_dir_num)*grid%norder*parallel%ldn, mpinfo)

!                write(9,*) ' Before scatterv solver%adiag '
!                call myflush(9)
!		call MPI_BCAST(adiag, parallel%ndim, MPI_DOUBLE_PRECISION, parallel%masterid, parallel%comm)
!		call MPI_SCATTERV(adiag, parallel%irows(1:parallel%procs_num)-parallel%irows(0:parallel%procs_num-1), &
!			parallel%irows(0:parallel%procs_num-1)-1, MPI_DOUBLE_PRECISION, solver%adiag, &
!			parallel%mydim, MPI_DOUBLE_PRECISION, parallel%group_master, parallel%group_comm, mpinfo)
		!solver%adiag(1:parallel%mydim) = adiag(parallel%irows(parallel%group_iam-1):(parallel%irows(parallel%group_iam)-1))
		! Set up kInverse, a version of grid%indexg() that actually is an inverse of grid%kx, ky, and kz  
		kInverse(:,:,:) = 0
		do ii = 1, parallel%nwedge
			kInverse(grid%kx(ii), grid%ky(ii), grid%kz(ii)) = ii  
		enddo 
!		write(9,*) ' kInverse = ', kInverse(1:3, 1:3, 1:3) 
		do ii = 1, parallel%nwedge  
			igrid = kInverse(kx(ii), ky(ii), kz(ii)) 
			if (igrid >= parallel%irows(parallel%group_iam) .and. igrid < parallel%irows(parallel%group_iam+1)) then 
				solver%adiag(igrid - parallel%irows(parallel%group_iam) + 1) = adiag(ii) 
!				write(9,*) ' solver%adiag(', igrid - parallel%irows(parallel%group_iam) + 1, ') = adiag( ', ii, ')' 
				do isp = 1, pot%nspin
					pot%vnew(igrid - parallel%irows(parallel%group_iam) + 1, isp) = vnew(ii, isp) 
				enddo 
			endif 
		enddo 
!        write(9,*) ' After scatterv solver%adiag '
!		call myflush(9)
		!if (allocated (adiag)) 	deallocate(adiag)
		! Set parallel%mydim based on irows, as in l. 343 of setup() 
		! Reset parallel%neibs to its global form (as in l. 157-161 and 187-294 of setup()), since this is what comm_neigh() expects on input.  
                call MPI_BARRIER(parallel%comm,mpinfo)

!                write(9,*) ' After comm_neigh() '
!                call myflush(9)

		call nonloc(clust,grid,p_pot,nloc_p_pot,pbc,parallel,mpinfo)
		call upot(clust,grid,p_pot,u_pot,pbc,parallel,elec_st%nspin, &
			elec_st%cplx,mpinfo)

                !write(9,*) ' Before set to zero: anloc(1,:) = ', nloc_p_pot%anloc(1,1:5)
                !write(9,*) ' Before set to zero: anloc(2,:) = ', nloc_p_pot%anloc(2,1:5)
                !call myflush(9)
		!nloc_p_pot%anloc(1:maxval(nloc_p_pot%nlatom), 1:(nloc_p_pot%nlm)) = 0.d0  
		!nloc_p_pot%anloc(1,3) = 0.d0 
		!nloc_p_pot%anloc(2,3) = 0.d0 
!		write(9,*) ' After set to zero: anloc(1,:) = ', nloc_p_pot%anloc(1,1:5)
!		write(9,*) ' After set to zero: anloc(2,:) = ', nloc_p_pot%anloc(2,1:5)
!		call myflush(9) 

	endif
	
	
  ! ===============================================================
  ! Assign representations, spins, and k-points round-robin across the blocks.
  ! If spin-orbital potentials are present, both spin components must
  ! be assigned the same block.
  ! ===============================================================
  ii = 0
  if (nloc_p_pot%is_so .or. elec_st%ncl) then
     do kplp = 1, max(elec_st%nkpt,1)

        elec_st%eig(1,kplp,1)%group = ii
        ii = ii + 1
        if (mod(ii,parallel%groups_num) == 0) ii = 0

!       AMIR - have to check this with DORON.
!        do irp = 1, elec_st%nrep
!           elec_st%eig(irp,kplp,:)%group = ii
!           ii = ii + 1
!           if (mod(ii,parallel%groups_num) == 0) ii = 0
!        enddo

     enddo
  else
     ! Determine how many k-points should be parallelized over the sign of k when R_z,pi symmetry is absent 
	 if ((.not. elec_st%cplx) .or. (elec_st%nkpt < 1) .or. (cond%RzSymm)) then
		cond%Parallelize_Over_SignK = -1
	 elseif (cond%Parallelize_Over_SignK >= 0) then 
		! First, count the total number of Arpack eigensolution computations that must be performed. 
		NumEigensolutionsPerSpin = 0
        do kplp = 1, elec_st%nkpt 
			if ( (cond%Num_Ignore_KPoints < 1 .or. &
			.not. any(cond%Ignore_KPoints(1:cond%Num_Ignore_KPoints) == kplp)) ) then
				! The Gamma point requires one Arpack eigensolution, all other k-points require two (for +-k) 
				! when R_z,pi symmetry is absent. 
				if( any(abs(elec_st%kpts(:,kplp)) > 5.d-5) ) then 
					NumEigensolutionsPerSpin = NumEigensolutionsPerSpin + 2 
				else
					NumEigensolutionsPerSpin = NumEigensolutionsPerSpin + 1 
				endif 
			endif 
		enddo 
		! Scale up NumEigensolutionsPerSpin for spins and symmetry group representations.  Note elec_st%nrep is 
		! expected to be 1, because ignore_symmetry should be true whenever conductance calculations are performed.  
		NumEigensolutionsPerSpin = max(NumEigensolutionsPerSpin,1) * elec_st%nspin * elec_st%nrep 
		! How many eigensolutions would remain after the maximum concurrent number are performed in 
		! parallel over k-points, but in serial over the sign of k 
		cond%Parallelize_Over_SignK = modulo(NumEigensolutionsPerSpin, 2*parallel%groups_num) 
		! If the remaining number cannot all be performed in parallel on the available number of MPI groups, 
		! parallelization over the sign of k would not save significant wall time 
		if (cond%Parallelize_Over_SignK <= 1 .or. &
			cond%Parallelize_Over_SignK > parallel%groups_num ) &
				cond%Parallelize_Over_SignK = -1 		
	 endif 
	 
	 ! Assign the transport calculations to MPI groups by representation, spin, and k-point 
	 NumEigensolutionsPerSpin = 0 
     do isp = 1, elec_st%nspin
        do kplp = 1, max(elec_st%nkpt,1)
           ! Note elec_st%nrep is expected to be 1, because ignore_symmetry should be true whenever conductance calculations 
		   ! are performed. 
		   do irp = 1, elec_st%nrep
              ! In the case that a spin-polarized calculation is parallelized over the sign of k, 
			  ! *both* spins of the gamma point should be computed in parallel, in order that the 
			  ! other k-points will be optimally load-balanced.  
			  if (isp > 1 .and. cond%Parallelize_Over_SignK > 0 .and. kplp == 1 .and. &
				NumEigensolutionsPerSpin >= cond%Parallelize_Over_SignK) then 
				if (.not. any(abs(elec_st%kpts(:,kplp)) > 5.d-5)) cycle 
			  endif 
			  ! Skip any k-points listed in the cond_skip_kpoints block in the input file 
              if ( (cond%Num_Ignore_KPoints < 1 .or. &
				.not. any(cond%Ignore_KPoints(1:cond%Num_Ignore_KPoints) == kplp)) ) then
					! Assign an MPI group   
					elec_st%eig(irp,kplp,isp)%group = ii
					ii = ii + 1
					! Deal with parallelization over the sign of k  
					if (NumEigensolutionsPerSpin < cond%Parallelize_Over_SignK .and. &
						NumEigensolutionsPerSpin < parallel%groups_num) then
						NumEigensolutionsPerSpin = NumEigensolutionsPerSpin + 1
						! For k-points besides the gamma point, assign the +k and -k points to separate MPI groups 
						! in parallel if this is among the first cond%Parallelize_Over_SignK eigensolutions. 
						! We don't modify elec_st%eig to keep track of this info, because eigensolution only needs to be 
						! performed for -k during conductance calculations, and only when R_z,pi symmetry is absent.  Instead, we 
						! follow the rule that the -k and +k eigensolutions are performed in MPI groups with successive group indices.  
						if ( any(abs(elec_st%kpts(:,kplp)) > 5.d-5) ) then 
							ii = ii + 1
							NumEigensolutionsPerSpin = NumEigensolutionsPerSpin + 1 
						! Without spin polarization, only one eigensolution is performed for the gamma point.  
						! (The conditional will fail because isp = 1 and nspin <= 1).  With spin polarization, both  
						! spins of the gamma point are done in parallel in order to optimize the overall load balancing.  
						elseif (isp < elec_st%nspin) then 
							elec_st%eig(irp,kplp,isp+1)%group = ii
							ii = ii + 1
							NumEigensolutionsPerSpin = NumEigensolutionsPerSpin + 1 
						endif 
					endif 
					!if (mod(ii,parallel%groups_num) == 0) ii = 0
					if (ii >= parallel%groups_num) ii = mod(ii,parallel%groups_num) 
				endif 
           enddo
        enddo
     enddo
  endif
#endif 

 endif 

 end subroutine Change_MPI_Gps_Cond 




